"""
Test cases for parser module
"""

import pytest
import os
from pathlib import Path
from unittest.mock import Mock, patch

from roseApp.core.parser import (
    ParserType, ParserHealth, IBagParser,
    RosbagsBagParser, LegacyBagParser, ParserHealthChecker,
    create_parser, create_best_parser, get_parser_health,
    get_all_parser_health, check_parser_availability,
    FileExistsError
)


class TestParserType:
    """Test cases for ParserType enum"""
    
    def test_parser_type_values(self):
        """Test parser type enum values"""
        assert ParserType.ROSBAGS.value == "rosbags"
        assert ParserType.LEGACY.value == "legacy"


class TestParserHealth:
    """Test cases for ParserHealth class"""
    
    def test_parser_health_creation(self):
        """Test basic ParserHealth creation"""
        health = ParserHealth(
            parser_type=ParserType.ROSBAGS,
            available=True,
            version="1.0.0",
            performance_score=100.0
        )
        
        assert health.parser_type == ParserType.ROSBAGS
        assert health.available
        assert health.version == "1.0.0"
        assert health.performance_score == 100.0
        assert health.is_healthy()
    
    def test_parser_health_unhealthy(self):
        """Test unhealthy parser health"""
        health = ParserHealth(
            parser_type=ParserType.ROSBAGS,
            available=False,
            error_message="Import failed"
        )
        
        assert not health.is_healthy()


class TestRosbagsBagParser:
    """Test cases for RosbagsBagParser"""
    
    def test_parser_initialization(self):
        """Test basic parser initialization"""
        parser = RosbagsBagParser()
        assert parser is not None
    
    def test_load_whitelist(self, temp_dir):
        """Test loading topics from whitelist file"""
        whitelist_path = temp_dir / "whitelist.txt"
        whitelist_path.write_text("/topic1\n/topic2\n# This is a comment\n/topic3\n")
        
        parser = RosbagsBagParser()
        topics = parser.load_whitelist(str(whitelist_path))
        
        assert len(topics) == 3
        assert "/topic1" in topics
        assert "/topic2" in topics
        assert "/topic3" in topics
    
    @patch('rosbags.highlevel.AnyReader')
    def test_load_bag(self, mock_any_reader, temp_dir):
        """Test loading bag file"""
        mock_reader = Mock()
        mock_reader.connections = [
            Mock(topic="/topic1", msgtype="std_msgs/String"),
            Mock(topic="/topic2", msgtype="geometry_msgs/Twist")
        ]
        mock_reader.start_time = 1234567890000000000
        mock_reader.end_time = 1234567900000000000
        mock_any_reader.return_value.__enter__.return_value = mock_reader
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = RosbagsBagParser()
        topics, connections, time_range = parser.load_bag(str(bag_path))
        
        assert len(topics) == 2
        assert "/topic1" in topics
        assert connections["/topic1"] == "std_msgs/String"
        assert time_range == ((123456789, 0), (123456790, 0))
    
    @patch('rosbags.highlevel.AnyReader')
    def test_get_message_counts(self, mock_any_reader, temp_dir):
        """Test getting message counts"""
        mock_reader = Mock()
        mock_connection1 = Mock(topic="/topic1")
        mock_connection2 = Mock(topic="/topic2")
        mock_reader.connections = [mock_connection1, mock_connection2]
        mock_reader.messages.side_effect = [
            [(None, None, None)] * 100,  # 100 messages for topic1
            [(None, None, None)] * 50    # 50 messages for topic2
        ]
        mock_any_reader.return_value.__enter__.return_value = mock_reader
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = RosbagsBagParser()
        counts = parser.get_message_counts(str(bag_path))
        
        assert counts["/topic1"] == 100
        assert counts["/topic2"] == 50
    
    @patch('rosbags.highlevel.AnyReader')
    def test_get_topic_sizes(self, mock_any_reader, temp_dir):
        """Test getting topic sizes"""
        mock_reader = Mock()
        mock_connection1 = Mock(topic="/topic1")
        mock_connection2 = Mock(topic="/topic2")
        mock_reader.connections = [mock_connection1, mock_connection2]
        mock_reader.messages.side_effect = [
            [(None, None, b"data" * 100)] * 10,  # 1000 bytes for topic1
            [(None, None, b"data" * 50)] * 5    # 250 bytes for topic2
        ]
        mock_any_reader.return_value.__enter__.return_value = mock_reader
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = RosbagsBagParser()
        sizes = parser.get_topic_sizes(str(bag_path))
        
        assert sizes["/topic1"] == 4000  # 10 * 400 bytes
        assert sizes["/topic2"] == 1000  # 5 * 200 bytes
    
    @patch('rosbags.highlevel.AnyReader')
    def test_get_topic_stats(self, mock_any_reader, temp_dir):
        """Test getting comprehensive topic statistics"""
        mock_reader = Mock()
        mock_connection = Mock(topic="/topic1")
        mock_reader.connections = [mock_connection]
        mock_reader.messages.return_value = [(None, None, b"data" * 100)] * 10
        mock_any_reader.return_value.__enter__.return_value = mock_reader
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = RosbagsBagParser()
        stats = parser.get_topic_stats(str(bag_path))
        
        assert "/topic1" in stats
        assert stats["/topic1"]["count"] == 10
        assert stats["/topic1"]["size"] == 4000
        assert stats["/topic1"]["avg_size"] == 400
    
    @patch('rosbags.highlevel.AnyReader')
    def test_read_messages(self, mock_any_reader, temp_dir):
        """Test reading messages from bag"""
        mock_reader = Mock()
        mock_connection = Mock(topic="/topic1", msgtype="std_msgs/String")
        mock_reader.connections = [mock_connection]
        mock_reader.messages.return_value = [
            (mock_connection, 1234567890000000000, b"message_data")
        ]
        mock_reader.deserialize.return_value = Mock(data="test_message")
        mock_any_reader.return_value.__enter__.return_value = mock_reader
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = RosbagsBagParser()
        messages = list(parser.read_messages(str(bag_path), ["/topic1"]))
        
        assert len(messages) == 1
        timestamp, message = messages[0]
        assert timestamp == (123456789, 0)
        assert message.data == "test_message"
    
    @patch('rosbags.rosbag1.Writer')
    @patch('rosbags.highlevel.AnyReader')
    def test_filter_bag(self, mock_any_reader, mock_writer, temp_dir):
        """Test filtering bag file"""
        # Mock AnyReader
        mock_reader = Mock()
        mock_connection = Mock(
            topic="/topic1",
            msgtype="std_msgs/String",
            msgdef="string data",
            digest="992ce8a1687cec8c8bd883ec73ca41d1"
        )
        mock_reader.connections = [mock_connection]
        mock_reader.messages.return_value = [
            (mock_connection, 1234567890000000000, b"message_data")
        ]
        mock_any_reader.return_value.__enter__.return_value = mock_reader
        
        # Mock Writer
        mock_writer_instance = Mock()
        mock_writer.return_value.__enter__.return_value = mock_writer_instance
        
        input_path = temp_dir / "input.bag"
        output_path = temp_dir / "output.bag"
        input_path.touch()
        
        parser = RosbagsBagParser()
        result = parser.filter_bag(
            str(input_path),
            str(output_path),
            ["/topic1"],
            compression="bz2"
        )
        
        assert "completed" in result.lower()
    
    def test_get_compression_format(self):
        """Test compression format conversion"""
        parser = RosbagsBagParser()
        
        with patch('rosbags.rosbag1.Writer') as mock_writer:
            mock_writer.CompressionFormat.BZ2 = "BZ2"
            mock_writer.CompressionFormat.LZ4 = "LZ4"
            
            assert parser._get_compression_format("bz2") == "BZ2"
            assert parser._get_compression_format("lz4") == "LZ4"
            assert parser._get_compression_format("none") is None
    
    def test_filter_bag_file_exists_error(self, temp_dir):
        """Test error when output file exists"""
        input_path = temp_dir / "input.bag"
        output_path = temp_dir / "output.bag"
        input_path.touch()
        output_path.touch()
        
        parser = RosbagsBagParser()
        
        with pytest.raises(FileExistsError, match="already exists"):
            parser.filter_bag(str(input_path), str(output_path), ["/topic1"])


class TestLegacyBagParser:
    """Test cases for LegacyBagParser"""
    
    def test_parser_initialization(self):
        """Test basic parser initialization"""
        parser = LegacyBagParser()
        assert parser is not None
    
    def test_load_whitelist(self, temp_dir):
        """Test loading topics from whitelist file"""
        whitelist_path = temp_dir / "whitelist.txt"
        whitelist_path.write_text("/topic1\n/topic2\n# This is a comment\n/topic3\n")
        
        parser = LegacyBagParser()
        topics = parser.load_whitelist(str(whitelist_path))
        
        assert len(topics) == 3
        assert "/topic1" in topics
        assert "/topic2" in topics
        assert "/topic3" in topics
    
    @patch('rosbag.Bag')
    def test_load_bag(self, mock_bag_class, temp_dir):
        """Test loading bag file with legacy parser"""
        mock_bag = Mock()
        mock_bag.get_type_and_topic_info.return_value = Mock(
            topics={
                "/topic1": Mock(msg_type="std_msgs/String"),
                "/topic2": Mock(msg_type="geometry_msgs/Twist")
            }
        )
        mock_bag.get_start_time.return_value = 123456789.0
        mock_bag.get_end_time.return_value = 123456790.0
        mock_bag_class.return_value.__enter__.return_value = mock_bag
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = LegacyBagParser()
        topics, connections, time_range = parser.load_bag(str(bag_path))
        
        assert len(topics) == 2
        assert "/topic1" in topics
        assert connections["/topic1"] == "std_msgs/String"
        assert time_range == ((123456789, 0), (123456790, 0))
    
    @patch('rosbag.Bag')
    def test_get_message_counts(self, mock_bag_class, temp_dir):
        """Test getting message counts with legacy parser"""
        mock_bag = Mock()
        mock_bag.get_type_and_topic_info.return_value = Mock(
            topics={
                "/topic1": Mock(message_count=100),
                "/topic2": Mock(message_count=50)
            }
        )
        mock_bag_class.return_value.__enter__.return_value = mock_bag
        
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        parser = LegacyBagParser()
        counts = parser.get_message_counts(str(bag_path))
        
        assert counts["/topic1"] == 100
        assert counts["/topic2"] == 50
    
    @patch('rosbag.Bag')
    def test_filter_bag(self, mock_bag_class, temp_dir):
        """Test filtering bag with legacy parser"""
        mock_input_bag = Mock()
        mock_output_bag = Mock()
        
        # Mock input bag messages
        mock_input_bag.read_messages.return_value = [
            ("/topic1", Mock(), Mock(secs=123456789, nsecs=0))
        ]
        
        mock_bag_class.side_effect = [
            mock_input_bag,  # Input bag
            mock_output_bag  # Output bag
        ]
        
        input_path = temp_dir / "input.bag"
        output_path = temp_dir / "output.bag"
        input_path.touch()
        
        parser = LegacyBagParser()
        result = parser.filter_bag(
            str(input_path),
            str(output_path),
            ["/topic1"],
            compression="bz2"
        )
        
        assert "completed" in result.lower()
        mock_output_bag.write.assert_called_once()


class TestParserHealthChecker:
    """Test cases for ParserHealthChecker"""
    
    def test_health_checker_initialization(self):
        """Test basic health checker initialization"""
        checker = ParserHealthChecker()
        assert checker is not None
    
    @patch('rosbags.highlevel.AnyReader')
    def test_check_rosbags_health_success(self, mock_any_reader):
        """Test rosbags health check success"""
        mock_any_reader.return_value = Mock()
        
        checker = ParserHealthChecker()
        health = checker._check_rosbags_health()
        
        assert health.parser_type == ParserType.ROSBAGS
        assert health.available
        assert health.performance_score == 100.0
    
    def test_check_rosbags_health_failure(self):
        """Test rosbags health check failure"""
        with patch.dict('sys.modules', {'rosbags': None}):
            checker = ParserHealthChecker()
            health = checker._check_rosbags_health()
            
            assert not health.available
            assert "rosbags not available" in health.error_message
    
    @patch('rosbag.Bag')
    def test_check_legacy_health_success(self, mock_bag):
        """Test legacy health check success"""
        checker = ParserHealthChecker()
        health = checker._check_legacy_health()
        
        assert health.parser_type == ParserType.LEGACY
        assert health.available
        assert health.performance_score == 30.0
    
    def test_check_legacy_health_failure(self):
        """Test legacy health check failure"""
        with patch.dict('sys.modules', {'rosbag': None}):
            checker = ParserHealthChecker()
            health = checker._check_legacy_health()
            
            assert not health.available
            assert "legacy rosbag not available" in health.error_message
    
    def test_get_best_parser(self):
        """Test getting best available parser"""
        checker = ParserHealthChecker()
        
        with patch.object(checker, 'check_parser_health') as mock_check:
            # Mock rosbags as healthy
            mock_check.return_value = ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=True
            )
            
            best_type = checker.get_best_parser()
            assert best_type == ParserType.ROSBAGS
    
    def test_get_best_parser_fallback(self):
        """Test fallback to legacy parser"""
        checker = ParserHealthChecker()
        
        with patch.object(checker, 'check_parser_health') as mock_check:
            def mock_health_check(parser_type):
                if parser_type == ParserType.ROSBAGS:
                    return ParserHealth(
                        parser_type=ParserType.ROSBAGS,
                        available=False,
                        error_message="rosbags not available"
                    )
                else:
                    return ParserHealth(
                        parser_type=ParserType.LEGACY,
                        available=True
                    )
            
            mock_check.side_effect = mock_health_check
            best_type = checker.get_best_parser()
            assert best_type == ParserType.LEGACY
    
    def test_get_all_health_status(self):
        """Test getting all parser health status"""
        checker = ParserHealthChecker()
        
        with patch.object(checker, 'check_parser_health') as mock_check:
            mock_check.return_value = ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=True
            )
            
            health_status = checker.get_all_health_status()
            assert ParserType.ROSBAGS in health_status
            assert ParserType.LEGACY in health_status


class TestGlobalFunctions:
    """Test cases for global functions"""
    
    def test_create_parser_rosbags(self):
        """Test creating rosbags parser"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.check_parser_health.return_value = ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=True
            )
            mock_checker.return_value = mock_instance
            
            parser = create_parser(ParserType.ROSBAGS)
            assert isinstance(parser, RosbagsBagParser)
    
    def test_create_parser_legacy(self):
        """Test creating legacy parser"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.check_parser_health.return_value = ParserHealth(
                parser_type=ParserType.LEGACY,
                available=True
            )
            mock_checker.return_value = mock_instance
            
            parser = create_parser(ParserType.LEGACY)
            assert isinstance(parser, LegacyBagParser)
    
    def test_create_parser_unhealthy(self):
        """Test creating parser with unhealthy status"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.check_parser_health.return_value = ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=False,
                error_message="Import failed"
            )
            mock_checker.return_value = mock_instance
            
            with pytest.raises(RuntimeError, match="not healthy"):
                create_parser(ParserType.ROSBAGS)
    
    def test_create_best_parser(self):
        """Test creating best available parser"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.get_best_parser.return_value = ParserType.ROSBAGS
            mock_instance.check_parser_health.return_value = ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=True
            )
            mock_checker.return_value = mock_instance
            
            parser = create_best_parser()
            assert isinstance(parser, RosbagsBagParser)
    
    def test_get_parser_health(self):
        """Test getting parser health"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.check_parser_health.return_value = ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=True
            )
            mock_checker.return_value = mock_instance
            
            health = get_parser_health(ParserType.ROSBAGS)
            assert health.parser_type == ParserType.ROSBAGS
            assert health.available
    
    def test_get_all_parser_health(self):
        """Test getting all parser health"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.get_all_health_status.return_value = {
                ParserType.ROSBAGS: ParserHealth(
                    parser_type=ParserType.ROSBAGS,
                    available=True
                ),
                ParserType.LEGACY: ParserHealth(
                    parser_type=ParserType.LEGACY,
                    available=True
                )
            }
            mock_checker.return_value = mock_instance
            
            health_status = get_all_parser_health()
            assert len(health_status) == 2
            assert ParserType.ROSBAGS in health_status
            assert ParserType.LEGACY in health_status
    
    def test_check_parser_availability(self):
        """Test checking parser availability"""
        with patch('roseApp.core.parser.ParserHealthChecker') as mock_checker:
            mock_instance = Mock()
            mock_instance.get_all_health_status.return_value = {
                ParserType.ROSBAGS: ParserHealth(
                    parser_type=ParserType.ROSBAGS,
                    available=True
                ),
                ParserType.LEGACY: ParserHealth(
                    parser_type=ParserType.LEGACY,
                    available=False
                )
            }
            mock_checker.return_value = mock_instance
            
            availability = check_parser_availability()
            assert availability["rosbags"] is True
            assert availability["legacy"] is False


#!/usr/bin/env python3
"""
Plot command for ROS bag data visualization
"""

import os
import time
from typing import Dict, List, Any, Optional, Union
from pathlib import Path
import typer
from rich.console import Console

# Import unified theme system
from ..core.theme import theme
from ..core.parser import create_parser, ParserType
from ..core.util import set_app_mode, AppMode, get_logger, log_cli_error
from .error_handling import ValidationError, validate_file_exists, validate_choice, validate_series_format, handle_runtime_error

try:
    import matplotlib
    matplotlib.use('Agg')  # Use non-interactive backend
    import matplotlib.pyplot as plt
    import matplotlib.dates as mdates
    from matplotlib.ticker import FuncFormatter
    MATPLOTLIB_AVAILABLE = True
except ImportError:
    MATPLOTLIB_AVAILABLE = False

try:
    import plotly.graph_objects as go
    import plotly.express as px
    from plotly.subplots import make_subplots
    import plotly.offline as pyo
    PLOTLY_AVAILABLE = True
except ImportError:
    PLOTLY_AVAILABLE = False

try:
    import pandas as pd
    PANDAS_AVAILABLE = True
except ImportError:
    PANDAS_AVAILABLE = False

try:
    import numpy as np
    NUMPY_AVAILABLE = True
except ImportError:
    NUMPY_AVAILABLE = False


app = typer.Typer(help="Generate data visualization plots from ROS bag files")


class PlottingError(Exception):
    """Exception raised for plotting-related errors"""
    pass


def check_plotting_dependencies():
    """Check if plotting dependencies are available"""
    missing = []
    if not MATPLOTLIB_AVAILABLE:
        missing.append("matplotlib")
    if not PLOTLY_AVAILABLE:
        missing.append("plotly")
    if not PANDAS_AVAILABLE:
        missing.append("pandas")
    
    if missing:
        missing_str = ", ".join(missing)
        install_cmd = f"pip install {' '.join(missing)}"
        error_msg = f"Missing plotting dependencies: {missing_str}. Install with: {install_cmd}"
        raise ValidationError(error_msg)


@app.command("plot")
def plot_cmd(
    bag_path: str = typer.Argument(..., help="Input bag file path"),
    series: List[str] = typer.Option(..., "--series", "-s", help="Plot series in format topic:field1,field2 (can be repeated)"),
    output: str = typer.Option(..., "--output", "-o", help="Output file path (required)"),
    plot_type: str = typer.Option("line", "--type", "-t", help="Plot type: line, scatter (default: line)"),
    as_format: str = typer.Option("png", "--as", "-a", help="Output format: png, svg, pdf, html (default: png)")
):
    """
    Generate data visualization plots from ROS bag files
    
    The --series parameter specifies what data to plot in the format:
    topic:field1,field2,...
    
    Examples:
    
    # Plot single field from one topic
    rose plot demo.bag --series /odom:pose.pose.position.x --output pos_x.png
    
    # Plot multiple fields from one topic  
    rose plot demo.bag --series /odom:pose.pose.position.x,pose.pose.position.y --output pos_xy.png
    
    # Plot all numeric fields from a topic
    rose plot demo.bag --series /odom: --output odom_all.png
    
    # Compare multiple topics
    rose plot demo.bag --series /odom:pose.pose.position.x --series /tf:transform.translation.x --output comparison.png
    
    # Generate scatter plot
    rose plot demo.bag --series /odom:pose.pose.position.x,pose.pose.position.y --type scatter --output scatter.png
    
    # Generate interactive HTML plot
    rose plot demo.bag --series /odom:pose.pose.position.x --as html --output interactive.html
    """
    # Set application mode for proper logging
    set_app_mode(AppMode.CLI)
    logger = get_logger()
    console = Console()
    
    try:
        # Check dependencies first
        check_plotting_dependencies()
        
        # Validate parameter values
        try:
            validate_file_exists(bag_path, "bag file")
            validate_choice(plot_type, ["line", "scatter"], "--type")
            validate_choice(as_format, ["png", "svg", "pdf", "html"], "--as")
            validate_series_format(series)
        except ValidationError as e:
            handle_runtime_error(e, "Parameter validation")
        
        # Parse series specifications (validation done by error handler)
        parsed_series = []
        for s in series:
            topic, fields_str = s.split(':', 1)
            
            # Parse fields - empty string means all numeric fields
            if fields_str.strip():
                fields = [f.strip() for f in fields_str.split(',')]
            else:
                fields = []  # Empty means all numeric fields
            
            parsed_series.append({
                'topic': topic,
                'fields': fields
            })
        
        # Create output directory if it doesn't exist
        output_path = Path(output)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        console.print("[cyan]Analyzing bag file and extracting time series data...[/cyan]")
        
        # Create parser and analyze bag
        parser = create_parser(ParserType.ROSBAGS)
        
        # Extract time series data
        time_series_data = _extract_time_series_data(parser, bag_path, parsed_series, console)
        
        # Create the plot
        console.print(f"[cyan]Creating {plot_type} plot...[/cyan]")
        _create_time_series_plot(time_series_data, output, plot_type, as_format, console)
        
        console.print(f"[green]✓ Plot saved to: {output}[/green]")
            
    except typer.Exit:
        # Re-raise typer.Exit cleanly
        raise
    except Exception as e:
        # Handle runtime errors without stack trace
        handle_runtime_error(e, "Plot generation")


def _format_bytes(bytes_val):
    """Format bytes value for display"""
    if bytes_val == 0:
        return "0 B"
    for unit in ['B', 'KB', 'MB', 'GB']:
        if bytes_val < 1024:
            return f"{bytes_val:.1f} {unit}"
        bytes_val /= 1024
    return f"{bytes_val:.1f} TB"


def _format_duration(seconds):
    """Format duration in seconds to human readable format"""
    if seconds < 60:
        return f"{seconds:.1f}s"
    elif seconds < 3600:
        return f"{seconds/60:.1f}m"
    else:
        return f"{seconds/3600:.1f}h"


def create_frequency_plot(json_data: Dict[str, Any], output_path: str, plot_format: str = "png"):
    """Create frequency bar plot for topics"""
    check_plotting_dependencies()
    
    topics_data = json_data['topics']
    summary = json_data['summary']
    
    # Filter topics with frequency data
    topics_with_freq = [t for t in topics_data if t['frequency'] is not None]
    
    if not topics_with_freq:
        raise PlottingError("No frequency data available. Use --verbose to analyze message frequencies.")
    
    # Sort by frequency
    topics_with_freq.sort(key=lambda x: x['frequency'], reverse=True)
    
    if plot_format == "html":
        return _create_frequency_plot_plotly(topics_with_freq, summary, output_path)
    else:
        return _create_frequency_plot_matplotlib(topics_with_freq, summary, output_path, plot_format)


def _create_frequency_plot_matplotlib(topics_data, summary, output_path, plot_format):
    """Create frequency plot using matplotlib with theme"""
    # Apply theme
    theme.apply_matplotlib_style()
    
    topics = [t['topic'] for t in topics_data]
    frequencies = [t['frequency'] for t in topics_data]
    
    # Truncate long topic names for display
    display_topics = [t[:30] + "..." if len(t) > 30 else t for t in topics]
    
    fig, ax = plt.subplots(figsize=(12, 8))
    bars = ax.bar(range(len(topics)), frequencies, color=theme.get_plot_color(0), alpha=0.8)
    
    ax.set_xlabel('Topics', fontsize=12)
    ax.set_ylabel('Frequency (Hz)', fontsize=12)
    ax.set_title(f'Topic Message Frequencies - {summary["file_name"]}', fontsize=14, fontweight='bold')
    ax.set_xticks(range(len(topics)))
    ax.set_xticklabels(display_topics, rotation=45, ha='right')
    
    # Add value labels on bars
    for bar, freq in zip(bars, frequencies):
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width()/2., height + max(frequencies)*0.01,
                f'{freq:.1f}', ha='center', va='bottom', color=theme.TEXT_PRIMARY, fontsize=10)
    
    # Customize grid
    ax.grid(True, alpha=0.3, linestyle='-', linewidth=0.5)
    ax.set_axisbelow(True)
    
    plt.tight_layout()
    plt.savefig(output_path, format=plot_format, dpi=300, bbox_inches='tight')
    plt.close()
    
    return output_path


def _create_frequency_plot_plotly(topics_data, summary, output_path):
    """Create frequency plot using plotly with theme"""
    topics = [t['topic'] for t in topics_data]
    frequencies = [t['frequency'] for t in topics_data]
    
    # Apply theme template
    template = theme.get_plotly_template()
    
    fig = go.Figure(data=[
        go.Bar(
            x=topics,
            y=frequencies,
            marker_color=theme.get_plot_color(0),
            text=[f'{f:.1f} Hz' for f in frequencies],
            textposition='auto',
            hovertemplate='<b>%{x}</b><br>Frequency: %{y:.1f} Hz<extra></extra>'
        )
    ])
    
    fig.update_layout(
        title={
            'text': f'Topic Message Frequencies - {summary["file_name"]}',
            'x': 0.5,
            'xanchor': 'center',
            'font': {'size': 18}
        },
        xaxis_title='Topics',
        yaxis_title='Frequency (Hz)',
        xaxis_tickangle=-45,
        showlegend=False,
        **template['layout']
    )
    
    pyo.plot(fig, filename=output_path, auto_open=False)
    return output_path


def create_size_distribution_plot(json_data: Dict[str, Any], output_path: str, plot_format: str = "png"):
    """Create size distribution plot for topics"""
    check_plotting_dependencies()
    
    topics_data = json_data['topics']
    summary = json_data['summary']
    
    # Filter topics with size data
    topics_with_size = [t for t in topics_data if t['size'] is not None and t['size'] > 0]
    
    if not topics_with_size:
        raise PlottingError("No size data available. Use --verbose to analyze message sizes.")
    
    # Sort by size
    topics_with_size.sort(key=lambda x: x['size'], reverse=True)
    
    if plot_format == "html":
        return _create_size_plot_plotly(topics_with_size, summary, output_path)
    else:
        return _create_size_plot_matplotlib(topics_with_size, summary, output_path, plot_format)


def _create_size_plot_matplotlib(topics_data, summary, output_path, plot_format):
    """Create size distribution plot using matplotlib with theme"""
    # Apply theme
    theme.apply_matplotlib_style()
    
    topics = [t['topic'] for t in topics_data]
    sizes = [t['size'] for t in topics_data]
    
    # Truncate long topic names for display
    display_topics = [t[:30] + "..." if len(t) > 30 else t for t in topics]
    
    fig, ax = plt.subplots(figsize=(12, 8))
    bars = ax.bar(range(len(topics)), sizes, color=theme.get_plot_color(2), alpha=0.8)
    
    ax.set_xlabel('Topics', fontsize=12)
    ax.set_ylabel('Total Size (Bytes)', fontsize=12)
    ax.set_title(f'Topic Message Sizes - {summary["file_name"]}', fontsize=14, fontweight='bold')
    ax.set_xticks(range(len(topics)))
    ax.set_xticklabels(display_topics, rotation=45, ha='right')
    
    # Format y-axis to show human readable sizes
    ax.yaxis.set_major_formatter(FuncFormatter(lambda x, p: _format_bytes(x)))
    
    # Add value labels on bars
    for bar, size in zip(bars, sizes):
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width()/2., height + max(sizes)*0.01,
                _format_bytes(size), ha='center', va='bottom', color=theme.TEXT_PRIMARY, fontsize=10)
    
    # Customize grid
    ax.grid(True, alpha=0.3, linestyle='-', linewidth=0.5)
    ax.set_axisbelow(True)
    
    plt.tight_layout()
    plt.savefig(output_path, format=plot_format, dpi=300, bbox_inches='tight')
    plt.close()
    
    return output_path


def _create_size_plot_plotly(topics_data, summary, output_path):
    """Create size distribution plot using plotly with theme"""
    topics = [t['topic'] for t in topics_data]
    sizes = [t['size'] for t in topics_data]
    size_labels = [_format_bytes(s) for s in sizes]
    
    # Apply theme template
    template = theme.get_plotly_template()
    
    fig = go.Figure(data=[
        go.Bar(
            x=topics,
            y=sizes,
            marker_color=theme.get_plot_color(2),
            text=size_labels,
            textposition='auto',
            hovertemplate='<b>%{x}</b><br>Size: %{text}<extra></extra>'
        )
    ])
    
    fig.update_layout(
        title={
            'text': f'Topic Message Sizes - {summary["file_name"]}',
            'x': 0.5,
            'xanchor': 'center',
            'font': {'size': 18}
        },
        xaxis_title='Topics',
        yaxis_title='Total Size (Bytes)',
        xaxis_tickangle=-45,
        showlegend=False,
        **template['layout']
    )
    
    # Format y-axis
    fig.update_yaxis(tickformat='.2s')
    
    pyo.plot(fig, filename=output_path, auto_open=False)
    return output_path


def create_message_count_plot(json_data: Dict[str, Any], output_path: str, plot_format: str = "png"):
    """Create message count plot for topics"""
    check_plotting_dependencies()
    
    topics_data = json_data['topics']
    summary = json_data['summary']
    
    # Filter topics with count data
    topics_with_count = [t for t in topics_data if t['count'] is not None and t['count'] > 0]
    
    if not topics_with_count:
        raise PlottingError("No message count data available. Use --verbose to analyze message counts.")
    
    # Sort by count
    topics_with_count.sort(key=lambda x: x['count'], reverse=True)
    
    if plot_format == "html":
        return _create_count_plot_plotly(topics_with_count, summary, output_path)
    else:
        return _create_count_plot_matplotlib(topics_with_count, summary, output_path, plot_format)


def _create_count_plot_matplotlib(topics_data, summary, output_path, plot_format):
    """Create message count plot using matplotlib with theme"""
    # Apply theme
    theme.apply_matplotlib_style()
    
    topics = [t['topic'] for t in topics_data]
    counts = [t['count'] for t in topics_data]
    
    # Truncate long topic names for display
    display_topics = [t[:30] + "..." if len(t) > 30 else t for t in topics]
    
    fig, ax = plt.subplots(figsize=(12, 8))
    bars = ax.bar(range(len(topics)), counts, color=theme.get_plot_color(3), alpha=0.8)
    
    ax.set_xlabel('Topics', fontsize=12)
    ax.set_ylabel('Message Count', fontsize=12)
    ax.set_title(f'Topic Message Counts - {summary["file_name"]}', fontsize=14, fontweight='bold')
    ax.set_xticks(range(len(topics)))
    ax.set_xticklabels(display_topics, rotation=45, ha='right')
    
    # Add value labels on bars
    for bar, count in zip(bars, counts):
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width()/2., height + max(counts)*0.01,
                f'{count:,}', ha='center', va='bottom', color=theme.TEXT_PRIMARY, fontsize=10)
    
    # Customize grid
    ax.grid(True, alpha=0.3, linestyle='-', linewidth=0.5)
    ax.set_axisbelow(True)
    
    plt.tight_layout()
    plt.savefig(output_path, format=plot_format, dpi=300, bbox_inches='tight')
    plt.close()
    
    return output_path


def _create_count_plot_plotly(topics_data, summary, output_path):
    """Create message count plot using plotly with theme"""
    topics = [t['topic'] for t in topics_data]
    counts = [t['count'] for t in topics_data]
    
    # Apply theme template
    template = theme.get_plotly_template()
    
    fig = go.Figure(data=[
        go.Bar(
            x=topics,
            y=counts,
            marker_color=theme.get_plot_color(3),
            text=[f'{c:,}' for c in counts],
            textposition='auto',
            hovertemplate='<b>%{x}</b><br>Count: %{y:,}<extra></extra>'
        )
    ])
    
    fig.update_layout(
        title={
            'text': f'Topic Message Counts - {summary["file_name"]}',
            'x': 0.5,
            'xanchor': 'center',
            'font': {'size': 18}
        },
        xaxis_title='Topics',
        yaxis_title='Message Count',
        xaxis_tickangle=-45,
        showlegend=False,
        **template['layout']
    )
    
    pyo.plot(fig, filename=output_path, auto_open=False)
    return output_path


def create_overview_plot(json_data: Dict[str, Any], output_path: str, plot_format: str = "png"):
    """Create overview plot with multiple metrics"""
    check_plotting_dependencies()
    
    topics_data = json_data['topics']
    summary = json_data['summary']
    
    # Filter topics with complete data
    complete_topics = [t for t in topics_data if 
                      t['count'] is not None and t['size'] is not None and t['frequency'] is not None
                      and t['count'] > 0]
    
    if not complete_topics:
        raise PlottingError("No complete data available. Use --verbose to analyze all metrics.")
    
    # Sort by total size
    complete_topics.sort(key=lambda x: x['size'], reverse=True)
    
    if plot_format == "html":
        return _create_overview_plot_plotly(complete_topics, summary, output_path)
    else:
        return _create_overview_plot_matplotlib(complete_topics, summary, output_path, plot_format)


def _create_overview_plot_matplotlib(topics_data, summary, output_path, plot_format):
    """Create overview plot using matplotlib with theme"""
    # Apply theme
    theme.apply_matplotlib_style()
    
    topics = [t['topic'] for t in topics_data]
    counts = [t['count'] for t in topics_data]
    sizes = [t['size'] for t in topics_data]
    frequencies = [t['frequency'] for t in topics_data]
    
    # Truncate long topic names for display
    display_topics = [t[:25] + "..." if len(t) > 25 else t for t in topics]
    
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(16, 12))
    
    # Message counts
    bars1 = ax1.bar(range(len(topics)), counts, color=theme.get_plot_color(3), alpha=0.8)
    ax1.set_title('Message Counts', fontsize=14, fontweight='bold')
    ax1.set_ylabel('Count', fontsize=12)
    ax1.set_xticks(range(len(topics)))
    ax1.set_xticklabels(display_topics, rotation=45, ha='right')
    ax1.grid(True, alpha=0.3, linestyle='-', linewidth=0.5)
    ax1.set_axisbelow(True)
    
    # Sizes
    bars2 = ax2.bar(range(len(topics)), sizes, color=theme.get_plot_color(2), alpha=0.8)
    ax2.set_title('Total Sizes', fontsize=14, fontweight='bold')
    ax2.set_ylabel('Size (Bytes)', fontsize=12)
    ax2.set_xticks(range(len(topics)))
    ax2.set_xticklabels(display_topics, rotation=45, ha='right')
    ax2.yaxis.set_major_formatter(FuncFormatter(lambda x, p: _format_bytes(x)))
    ax2.grid(True, alpha=0.3, linestyle='-', linewidth=0.5)
    ax2.set_axisbelow(True)
    
    # Frequencies
    bars3 = ax3.bar(range(len(topics)), frequencies, color=theme.get_plot_color(0), alpha=0.8)
    ax3.set_title('Frequencies', fontsize=14, fontweight='bold')
    ax3.set_ylabel('Frequency (Hz)', fontsize=12)
    ax3.set_xticks(range(len(topics)))
    ax3.set_xticklabels(display_topics, rotation=45, ha='right')
    ax3.grid(True, alpha=0.3, linestyle='-', linewidth=0.5)
    ax3.set_axisbelow(True)
    
    # Summary stats
    ax4.axis('off')
    stats_text = f"""
Bag File: {summary['file_name']}
Total Topics: {summary['topic_count']}
Total Messages: {summary['total_messages']:,}
File Size: {summary['file_size_formatted']}
Duration: {summary['duration_formatted']}
Avg Rate: {summary['avg_rate_formatted']}
Theme: {'Dark' if theme.dark_mode else 'Light'} mode
    """.strip()
    ax4.text(0.1, 0.5, stats_text, transform=ax4.transAxes, fontsize=12,
             verticalalignment='center', color=theme.TEXT_PRIMARY,
             bbox=dict(boxstyle='round', facecolor=theme.SURFACE, alpha=0.8))
    ax4.set_title('Summary Statistics', fontsize=14, fontweight='bold')
    
    plt.suptitle(f'ROS Bag Overview - {summary["file_name"]}', fontsize=16, fontweight='bold')
    plt.tight_layout()
    plt.savefig(output_path, format=plot_format, dpi=300, bbox_inches='tight')
    plt.close()
    
    return output_path


def _create_overview_plot_plotly(topics_data, summary, output_path):
    """Create overview plot using plotly with theme"""
    topics = [t['topic'] for t in topics_data]
    counts = [t['count'] for t in topics_data]
    sizes = [t['size'] for t in topics_data]
    frequencies = [t['frequency'] for t in topics_data]
    
    # Apply theme template
    template = theme.get_plotly_template()
    
    fig = make_subplots(
        rows=2, cols=2,
        subplot_titles=('Message Counts', 'Total Sizes', 'Frequencies', 'Summary'),
        specs=[[{"type": "bar"}, {"type": "bar"}],
               [{"type": "bar"}, {"type": "table"}]]
    )
    
    # Message counts
    fig.add_trace(
        go.Bar(
            x=topics, 
            y=counts, 
            name='Count', 
            marker_color=theme.get_plot_color(3),
            hovertemplate='<b>%{x}</b><br>Count: %{y:,}<extra></extra>'
        ),
        row=1, col=1
    )
    
    # Sizes
    fig.add_trace(
        go.Bar(
            x=topics, 
            y=sizes, 
            name='Size', 
            marker_color=theme.get_plot_color(2),
            hovertemplate='<b>%{x}</b><br>Size: %{y}<extra></extra>'
        ),
        row=1, col=2
    )
    
    # Frequencies
    fig.add_trace(
        go.Bar(
            x=topics, 
            y=frequencies, 
            name='Frequency', 
            marker_color=theme.get_plot_color(0),
            hovertemplate='<b>%{x}</b><br>Frequency: %{y:.1f} Hz<extra></extra>'
        ),
        row=2, col=1
    )
    
    # Summary table
    fig.add_trace(
        go.Table(
            header=dict(
                values=['Metric', 'Value'],
                fill_color=theme.SURFACE,
                font=dict(color=theme.TEXT_PRIMARY, size=14),
                align='left'
            ),
            cells=dict(
                values=[
                    ['File', 'Topics', 'Messages', 'File Size', 'Duration', 'Avg Rate', 'Theme'],
                    [
                        summary['file_name'], 
                        summary['topic_count'], 
                        f"{summary['total_messages']:,}",
                        summary['file_size_formatted'], 
                        summary['duration_formatted'], 
                        summary['avg_rate_formatted'],
                        f"{'Dark' if theme.dark_mode else 'Light'} mode"
                    ]
                ],
                fill_color=theme.BACKGROUND,
                font=dict(color=theme.TEXT_PRIMARY, size=12),
                align='left'
            )
        ),
        row=2, col=2
    )
    
    fig.update_layout(
        title={
            'text': f"ROS Bag Overview - {summary['file_name']}",
            'x': 0.5,
            'xanchor': 'center',
            'font': {'size': 20}
        },
        showlegend=False,
        height=800,
        **template['layout']
    )
    
    # Update x-axes for better readability
    fig.update_xaxes(tickangle=-45, row=1, col=1)
    fig.update_xaxes(tickangle=-45, row=1, col=2)
    fig.update_xaxes(tickangle=-45, row=2, col=1)
    
    pyo.plot(fig, filename=output_path, auto_open=False)
    return output_path


def _extract_time_series_data(parser, bag_path: str, parsed_series: List[Dict], console: Console) -> Dict[str, Any]:
    """Extract time series data from bag file based on specified series"""
    import time
    from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn, TimeElapsedColumn
    
    time_series_data = {}
    
    # Load bag file
    topics, connections, time_range = parser.load_bag(bag_path)
    
    # Check if requested topics exist
    available_topics = set(topics)
    requested_topics = {series['topic'] for series in parsed_series}
    missing_topics = requested_topics - available_topics
    
    if missing_topics:
        console.print(f"[yellow]Warning: The following topics were not found in the bag file:[/yellow]")
        for topic in missing_topics:
            console.print(f"  • {topic}")
        console.print(f"[dim]Available topics: {', '.join(sorted(available_topics))}[/dim]")
    
    # Extract data for each series
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeElapsedColumn(),
        console=console
    ) as progress:
        
        for series in parsed_series:
            topic = series['topic']
            fields = series['fields']
            
            if topic not in available_topics:
                continue
                
            task = progress.add_task(f"Extracting {topic}", total=100)
            
            # Get messages for this topic
            messages = parser.read_messages(bag_path, [topic])
            
            topic_data = {
                'topic': topic,
                'fields': fields,
                'timestamps': [],
                'data': {}
            }
            
            # Initialize field data containers
            if fields:
                for field in fields:
                    topic_data['data'][field] = []
            
            message_count = 0
            processed_count = 0
            
            # Process messages directly (we can't easily count them first)
            for msg_timestamp, msg_data in messages:
                processed_count += 1
                
                # Update progress (use processed count as approximation)
                if processed_count % 10 == 0:  # Update every 10 messages
                    progress.update(task, completed=min(processed_count, 100))
                
                # Extract timestamp - convert from (seconds, nanoseconds) to float
                if isinstance(msg_timestamp, tuple):
                    timestamp = msg_timestamp[0] + msg_timestamp[1] / 1_000_000_000
                else:
                    timestamp = float(msg_timestamp)
                
                topic_data['timestamps'].append(timestamp)
                
                # Extract field data
                if fields:
                    for field in fields:
                        value = _extract_field_value(msg_data, field)
                        topic_data['data'][field].append(value)
                else:
                    # If no fields specified, try to extract all numeric fields
                    numeric_fields = _extract_numeric_fields(msg_data)
                    for field_name, value in numeric_fields.items():
                        if field_name not in topic_data['data']:
                            topic_data['data'][field_name] = []
                        topic_data['data'][field_name].append(value)
            
            progress.update(task, completed=100)
            time_series_data[topic] = topic_data
    
    return time_series_data


def _extract_field_value(msg_data: Any, field_path: str) -> float:
    """Extract a numeric value from a message using dot notation field path"""
    try:
        # Handle nested field access like "pose.pose.position.x"
        value = msg_data
        for field_name in field_path.split('.'):
            if hasattr(value, field_name):
                value = getattr(value, field_name)
            elif isinstance(value, dict) and field_name in value:
                value = value[field_name]
            else:
                return 0.0
        
        # Convert to float if possible
        if isinstance(value, (int, float)):
            return float(value)
        else:
            return 0.0
    except Exception:
        return 0.0


def _extract_numeric_fields(msg_data: Any, prefix: str = "", max_depth: int = 3) -> Dict[str, float]:
    """Extract all numeric fields from a message"""
    numeric_fields = {}
    
    if max_depth <= 0:
        return numeric_fields
    
    try:
        # Handle different message types
        if hasattr(msg_data, '__dict__'):
            # ROS message object
            for field_name in dir(msg_data):
                if field_name.startswith('_'):
                    continue
                    
                field_value = getattr(msg_data, field_name)
                full_name = f"{prefix}.{field_name}" if prefix else field_name
                
                if isinstance(field_value, (int, float)):
                    numeric_fields[full_name] = float(field_value)
                elif hasattr(field_value, '__dict__'):
                    # Nested object
                    nested_fields = _extract_numeric_fields(field_value, full_name, max_depth - 1)
                    numeric_fields.update(nested_fields)
        
        elif isinstance(msg_data, dict):
            # Dictionary
            for field_name, field_value in msg_data.items():
                full_name = f"{prefix}.{field_name}" if prefix else field_name
                
                if isinstance(field_value, (int, float)):
                    numeric_fields[full_name] = float(field_value)
                elif isinstance(field_value, dict):
                    nested_fields = _extract_numeric_fields(field_value, full_name, max_depth - 1)
                    numeric_fields.update(nested_fields)
    
    except Exception:
        pass
    
    return numeric_fields


def _create_time_series_plot(time_series_data: Dict[str, Any], output_path: str, plot_type: str, plot_format: str, console: Console):
    """Create time series plot using matplotlib or plotly"""
    import datetime
    
    if plot_format == "html":
        _create_time_series_plot_plotly(time_series_data, output_path, plot_type, console)
    else:
        _create_time_series_plot_matplotlib(time_series_data, output_path, plot_type, plot_format, console)


def _create_time_series_plot_matplotlib(time_series_data: Dict[str, Any], output_path: str, plot_type: str, plot_format: str, console: Console):
    """Create time series plot using matplotlib"""
    if not MATPLOTLIB_AVAILABLE:
        raise ValidationError("Missing matplotlib. Install with: pip install matplotlib")
    
    import matplotlib.pyplot as plt
    import matplotlib.dates as mdates
    from datetime import datetime, timezone
    
    # Create figure and axis
    fig, ax = plt.subplots(figsize=(12, 8))
    
    # Plot each topic's data
    for topic, data in time_series_data.items():
        timestamps = data['timestamps']
        
        # Convert timestamps to datetime objects
        datetime_stamps = [datetime.fromtimestamp(ts, tz=timezone.utc) for ts in timestamps]
        
        # Plot each field
        for field_name, field_data in data['data'].items():
            if len(field_data) != len(datetime_stamps):
                continue
                
            label = f"{topic}:{field_name}"
            
            if plot_type == "scatter":
                ax.scatter(datetime_stamps, field_data, label=label, alpha=0.6)
            else:  # line plot
                ax.plot(datetime_stamps, field_data, label=label, linewidth=1.5)
    
    # Customize plot
    ax.set_xlabel('Time')
    ax.set_ylabel('Value')
    ax.set_title('Time Series Plot')
    ax.legend()
    ax.grid(True, alpha=0.3)
    
    # Format x-axis
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M:%S'))
    ax.xaxis.set_major_locator(mdates.SecondLocator(interval=60))
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    
    # Adjust layout
    plt.tight_layout()
    
    # Save plot
    plt.savefig(output_path, format=plot_format, dpi=300, bbox_inches='tight')
    plt.close()


def _create_time_series_plot_plotly(time_series_data: Dict[str, Any], output_path: str, plot_type: str, console: Console):
    """Create time series plot using plotly"""
    if not PLOTLY_AVAILABLE:
        raise ValidationError("Missing plotly. Install with: pip install plotly")
    
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    from datetime import datetime, timezone
    
    # Create figure
    fig = go.Figure()
    
    # Plot each topic's data
    for topic, data in time_series_data.items():
        timestamps = data['timestamps']
        
        # Convert timestamps to datetime objects
        datetime_stamps = [datetime.fromtimestamp(ts, tz=timezone.utc) for ts in timestamps]
        
        # Plot each field
        for field_name, field_data in data['data'].items():
            if len(field_data) != len(datetime_stamps):
                continue
                
            name = f"{topic}:{field_name}"
            
            if plot_type == "scatter":
                fig.add_trace(go.Scatter(
                    x=datetime_stamps,
                    y=field_data,
                    mode='markers',
                    name=name,
                    opacity=0.6
                ))
            else:  # line plot
                fig.add_trace(go.Scatter(
                    x=datetime_stamps,
                    y=field_data,
                    mode='lines',
                    name=name,
                    line=dict(width=2)
                ))
    
    # Customize layout
    fig.update_layout(
        title='Time Series Plot',
        xaxis_title='Time',
        yaxis_title='Value',
        hovermode='x unified',
        showlegend=True,
        width=1200,
        height=600
    )
    
    # Save plot
    fig.write_html(output_path)


def create_plot(json_data: Dict[str, Any], plot_type: str, output_path: str, plot_format: str = "png"):
    """Create plot based on type"""
    plot_functions = {
        'frequency': create_frequency_plot,
        'size': create_size_distribution_plot,
        'count': create_message_count_plot,
        'overview': create_overview_plot
    }
    
    if plot_type not in plot_functions:
        raise PlottingError(f"Unknown plot type: {plot_type}. Available types: {', '.join(plot_functions.keys())}")
    
    return plot_functions[plot_type](json_data, output_path, plot_format)


def main():
    """Main entry point"""
    app()


if __name__ == "__main__":
    main() 


#!/usr/bin/env python3
"""
Inspect command for ROS bag files - Using ResultHandler for rendering and export
"""
import asyncio
from pathlib import Path
from typing import Optional, List

import typer
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

from ..core.bag_manager import BagManager, InspectOptions
from ..core.result_handler import OutputFormat, RenderOptions, ExportOptions

app = typer.Typer(help="Inspect ROS bag files")
console = Console()


@app.command()
def inspect(
    bag_path: Path = typer.Argument(..., help="Path to the ROS bag file"),
    topics: Optional[List[str]] = typer.Option(None, "--topics", "-t", help="Filter specific topics"),
    topic_filter: Optional[str] = typer.Option(None, "--filter", "-f", help="Filter topics by pattern"),
    show_fields: bool = typer.Option(False, "--show-fields", help="Show field analysis for messages"),
    sort_by: str = typer.Option("name", "--sort", help="Sort topics by (name, count, frequency)"),
    reverse_sort: bool = typer.Option(False, "--reverse", help="Reverse sort order"),
    limit: Optional[int] = typer.Option(None, "--limit", "-l", help="Limit number of topics shown"),
    as_format: str = typer.Option("table", "--as", help="Output format (table, list, summary, json, yaml, csv, xml, html, markdown)"),
    output: Optional[Path] = typer.Option(None, "--output", "-o", help="Output file path"),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output")
):
    """
    Inspect a ROS bag file and display comprehensive analysis
    
    This command uses the unified ResultHandler for all rendering and export operations.
    """
    # Validate bag file exists
    if not bag_path.exists():
        console.print(f"[red]Error: Bag file not found: {bag_path}[/red]")
        raise typer.Exit(1)
    
    # Convert string format to enum
    try:
        output_format = OutputFormat(as_format.lower())
    except ValueError:
        supported = ", ".join([fmt.value for fmt in OutputFormat])
        console.print(f"[red]Error: Unsupported output format '{as_format}'. Supported: {supported}[/red]")
        raise typer.Exit(1)
    
    # Create options object
    options = InspectOptions(
        topics=topics,
        topic_filter=topic_filter,
        show_fields=show_fields,
        sort_by=sort_by,
        reverse_sort=reverse_sort,
        limit=limit,
        output_format=output_format,
        output_file=output,
        verbose=verbose
    )
    
    # Run the async inspection
    asyncio.run(_run_inspect(bag_path, options))


async def _run_inspect(bag_path: Path, options: InspectOptions):
    """Run the bag inspection asynchronously using BagManager and ResultHandler"""
    
    # Create BagManager
    manager = BagManager()
    
    try:
        # Show progress during analysis
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            console=console,
            transient=True
        ) as progress:
            
            # Analysis phase
            progress.add_task("Analyzing bag file...", total=None)
            result = await manager.inspect_bag(bag_path, options)
            
            # Field analysis phase (if requested)
            if options.show_fields:
                progress.add_task("Analyzing field information...", total=None)
        
        # Get result handler
        handler = manager.get_result_handler()
        
        # Determine if we should export to file or render to console
        if options.output_file:
            # Export to file
            export_options = ExportOptions(
                format=options.output_format,
                output_file=options.output_file,
                pretty=True,
                include_metadata=True
            )
            
            success = handler.export(result, export_options)
            if not success:
                console.print("[red]Export failed[/red]")
                raise typer.Exit(1)
        else:
            # Render to console
            render_options = RenderOptions(
                format=options.output_format,
                verbose=options.verbose,
                show_fields=options.show_fields,
                show_cache_stats=True,
                show_summary=True,
                color=True,
                title=f"Topics in {bag_path.name}"
            )
            
            handler.render(result, render_options)
            
    except Exception as e:
        console.print(f"[red]Error: {e}[/red]")
        raise typer.Exit(1)
    finally:
        # Clean up resources
        manager.cleanup()


if __name__ == "__main__":
    app() 


"""
Enhanced ROS bag parser module with intelligent parser management.

This module provides comprehensive bag parsing capabilities with automatic
parser selection, health checking, and fallback mechanisms.
"""

import os
import time
import threading
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import List, Dict, Tuple, Optional, Callable, Any
from dataclasses import dataclass

from roseApp.core.util import get_logger

_logger = get_logger("parser")


class FileExistsError(Exception):
    """Custom exception for file existence errors"""
    pass


class ParserType(Enum):
    """Enum for different parser implementations"""
    ROSBAGS = "rosbags"  # Enhanced rosbags-based implementation (default)
    LEGACY = "legacy"    # Legacy rosbag implementation (fallback)


@dataclass
class ParserHealth:
    """Parser health status information"""
    parser_type: ParserType
    available: bool
    version: Optional[str] = None
    performance_score: float = 0.0
    last_check: float = 0.0
    error_message: Optional[str] = None
    
    def is_healthy(self) -> bool:
        """Check if parser is healthy and available"""
        return self.available and not self.error_message


class IBagParser(ABC):
    """Abstract base class for bag parser implementations"""
    
    @abstractmethod
    def load_whitelist(self, whitelist_path: str) -> List[str]:
        """Load topics from whitelist file"""
        pass
    
    @abstractmethod
    def filter_bag(self, input_bag: str, output_bag: str, topics: List[str], 
                  time_range: Optional[Tuple] = None, 
                  progress_callback: Optional[Callable] = None,
                  compression: str = 'none',
                  overwrite: bool = False) -> str:
        """Filter rosbag using selected implementation"""
        pass
    
    @abstractmethod
    def load_bag(self, bag_path: str) -> Tuple[List[str], Dict[str, str], Tuple]:
        """Load bag file and return topics, connections and time range"""
        pass
    
    @abstractmethod
    def inspect_bag(self, bag_path: str) -> str:
        """List all topics and message types"""
        pass

    @abstractmethod
    def get_message_counts(self, bag_path: str) -> Dict[str, int]:
        """Get message counts for each topic in the bag file"""
        pass

    @abstractmethod
    def get_topic_sizes(self, bag_path: str) -> Dict[str, int]:
        """Get total size in bytes for each topic in the bag file"""
        pass

    @abstractmethod
    def get_topic_stats(self, bag_path: str) -> Dict[str, Dict[str, int]]:
        """Get comprehensive statistics for each topic"""
        pass

    @abstractmethod
    def read_messages(self, bag_path: str, topics: List[str]):
        """Read messages from specified topics in the bag file"""
        pass


class RosbagsBagParser(IBagParser):
    """High-performance rosbags implementation using AnyReader/Rosbag1Writer"""
    
    def __init__(self):
        """Initialize enhanced rosbags parser"""
        self._registered_types = set()
        _logger.debug("Initialized RosbagsBagParser with enhanced performance features")
    
    def load_whitelist(self, whitelist_path: str) -> List[str]:
        """Load topics from whitelist file"""
        with open(whitelist_path) as f:
            topics = []
            for line in f.readlines():
                if line.strip() and not line.strip().startswith('#'):
                    topics.append(line.strip())
            return topics
    
    def filter_bag(self, input_bag: str, output_bag: str, topics: List[str], 
                  time_range: Optional[Tuple] = None,
                  progress_callback: Optional[Callable] = None,
                  compression: str = 'none',
                  overwrite: bool = False) -> str:
        """Filter bag file by topics and time range using AnyReader for performance"""
        try:
            # Validate compression type before starting
            from roseApp.core.util import validate_compression_type
            is_valid, error_message = validate_compression_type(compression)
            if not is_valid:
                raise ValueError(error_message)
            
            # Check if output file exists
            if os.path.exists(output_bag) and not overwrite:
                raise FileExistsError(f"Output file '{output_bag}' already exists. Use overwrite=True to overwrite.")
            
            # Remove existing file if overwrite is True
            if os.path.exists(output_bag) and overwrite:
                os.remove(output_bag)
            
            start_time = time.time()
            
            # Convert compression format for rosbags
            rosbags_compression = self._get_compression_format(compression)
            
            # Initialize progress tracking variables
            last_progress = -1
            processed_messages = 0
            
            # Count total messages first for progress tracking
            total_messages = 0
            selected_topic_counts = {}
            
            # Use AnyReader for enhanced performance
            from rosbags.highlevel import AnyReader
            from rosbags.rosbag1 import Writer as Rosbag1Writer
            
            with AnyReader([Path(input_bag)]) as reader:
                # Pre-filter connections based on selected topics
                selected_connections = [
                    conn for conn in reader.connections 
                    if conn.topic in topics
                ]
                
                if not selected_connections:
                    _logger.warning(f"No matching topics found in {input_bag}")
                    if progress_callback:
                        progress_callback(100)
                    return "No messages found for selected topics"
                
                # Count messages for each selected topic
                for connection in selected_connections:
                    # Use efficient message counting
                    count = sum(1 for _ in reader.messages([connection]))
                    selected_topic_counts[connection.topic] = count
                    total_messages += count
                
                if total_messages == 0:
                    _logger.warning(f"No messages found for selected topics in {input_bag}")
                    if progress_callback:
                        progress_callback(100)
                    return "No messages found for selected topics"
                
                # Create output directory if needed
                output_dir = os.path.dirname(output_bag)
                if output_dir:
                    os.makedirs(output_dir, exist_ok=True)
                
                # Filter and write messages using Rosbag1Writer
                output_path = Path(output_bag)
                writer = Rosbag1Writer(output_path)
                
                # Set compression if specified
                if rosbags_compression:
                    writer.set_compression(rosbags_compression)
                
                with writer:
                    # Convert time range if provided
                    start_ns = None
                    end_ns = None
                    if time_range:
                        start_ns = time_range[0][0] * 1_000_000_000 + time_range[0][1]
                        end_ns = time_range[1][0] * 1_000_000_000 + time_range[1][1]
                    
                    # Add connections to writer
                    topic_connections = {}
                    for connection in selected_connections:
                        # Extract connection information with proper defaults
                        callerid = '/rosbags_enhanced_parser'
                        if hasattr(connection, 'ext') and hasattr(connection.ext, 'callerid'):
                            if connection.ext.callerid is not None:
                                callerid = connection.ext.callerid
                        
                        msgdef = getattr(connection, 'msgdef', None)
                        md5sum = getattr(connection, 'digest', None)
                        
                        new_connection = writer.add_connection(
                            topic=connection.topic,
                            msgtype=connection.msgtype,
                            msgdef=msgdef,
                            md5sum=md5sum,
                            callerid=callerid
                        )
                        topic_connections[connection.topic] = new_connection
                    
                    # Process messages with progress tracking
                    for (connection, timestamp, rawdata) in reader.messages(connections=selected_connections):
                        # Check time range if specified
                        if time_range:
                            if timestamp < start_ns or timestamp > end_ns:
                                continue
                        
                        # Write message using connection mapping
                        writer.write(topic_connections[connection.topic], timestamp, rawdata)
                        
                        # Update progress
                        processed_messages += 1
                        if progress_callback and total_messages > 0:
                            current_progress = int((processed_messages / total_messages) * 100)
                            if current_progress != last_progress:
                                progress_callback(current_progress)
                                last_progress = current_progress
            
            end_time = time.time()
            elapsed = end_time - start_time
            mins, secs = divmod(elapsed, 60)
            
            if progress_callback and last_progress < 100:
                progress_callback(100)
                
            # Log performance statistics
            _logger.info(f"Filtered {processed_messages} messages from {len(selected_connections)} topics in {elapsed:.2f}s")
                
            return f"Filtering completed in {int(mins)}m {secs:.2f}s"
            
        except ValueError as ve:
            raise ve
        except FileExistsError as fe:
            raise fe
        except Exception as e:
            _logger.error(f"Error filtering bag with AnyReader: {e}")
            raise Exception(f"Error filtering bag: {e}")
    
    def _get_compression_format(self, compression: str):
        """Get rosbags CompressionFormat enum from string"""
        try:
            from rosbags.rosbag1 import Writer as Rosbag1Writer
            if compression == 'bz2':
                return Rosbag1Writer.CompressionFormat.BZ2
            elif compression == 'lz4':
                return Rosbag1Writer.CompressionFormat.LZ4
            else:
                return None
        except Exception:
            return None
    
    def load_bag(self, bag_path: str) -> Tuple[List[str], Dict[str, str], Tuple]:
        """Load bag file and return topics, connections and time range using AnyReader"""
        try:
            from rosbags.highlevel import AnyReader
            
            with AnyReader([Path(bag_path)]) as reader:
                # Get topics and message types
                topics = [conn.topic for conn in reader.connections]
                connections = {conn.topic: conn.msgtype for conn in reader.connections}
                
                # Get time range (AnyReader provides nanosecond timestamps)
                start_ns = reader.start_time
                end_ns = reader.end_time
                
                # Convert nanoseconds to (seconds, nanoseconds)
                start = (int(start_ns // 1_000_000_000), int(start_ns % 1_000_000_000))
                end = (int(end_ns // 1_000_000_000), int(end_ns % 1_000_000_000))
                
                return topics, connections, (start, end)
                
        except Exception as e:
            _logger.error(f"Error loading bag with AnyReader: {e}")
            raise Exception(f"Error loading bag: {e}")
    
    def inspect_bag(self, bag_path: str) -> str:
        """List all topics and message types in the bag file using AnyReader"""
        try:
            topics, connections, (start_time, end_time) = self.load_bag(bag_path)
            
            # Get topic statistics
            topic_stats = self.get_topic_stats(bag_path)
            
            # Helper function to format size
            def format_size(size_bytes: int) -> str:
                """Format size in bytes to human readable format"""
                size = float(size_bytes)
                for unit in ['B', 'KB', 'MB', 'GB']:
                    if size < 1024:
                        return f"{size:.1f}{unit}"
                    size /= 1024
                return f"{size:.1f}TB"
            
            from roseApp.core.util import TimeUtil
            
            result = [f"\nTopics in {bag_path}:"]
            result.append("{:<35} {:<35} {:<10} {:<10}".format("Topic", "Message Type", "Count", "Size"))
            result.append("-" * 90)
            
            for topic in topics:
                stats = topic_stats.get(topic, {'count': 0, 'size': 0})
                count = stats['count']
                size = stats['size']
                
                result.append("{:<35} {:<35} {:<10} {:<10}".format(
                    topic[:33], 
                    connections[topic][:33], 
                    count, 
                    format_size(size)
                ))
            
            # Calculate totals
            total_count = sum(stats['count'] for stats in topic_stats.values())
            total_size = sum(stats['size'] for stats in topic_stats.values())
            
            result.append("-" * 90)
            result.append(f"Total: {len(topics)} topics, {total_count} messages, {format_size(total_size)}")
            result.append(f"\nTime range: {TimeUtil.to_datetime(start_time)} - {TimeUtil.to_datetime(end_time)}")
            return "\n".join(result)
            
        except Exception as e:
            _logger.error(f"Error inspecting bag file: {e}")
            raise Exception(f"Error inspecting bag file: {e}")

    def get_message_counts(self, bag_path: str) -> Dict[str, int]:
        """Get message counts for each topic in the bag file using AnyReader"""
        try:
            from rosbags.highlevel import AnyReader
            
            with AnyReader([Path(bag_path)]) as reader:
                # Count messages for each topic
                topic_counts = {}
                
                for connection in reader.connections:
                    # Count messages efficiently
                    count = sum(1 for _ in reader.messages([connection]))
                    topic_counts[connection.topic] = count
                
                return topic_counts
            
        except Exception as e:
            _logger.error(f"Error getting message counts: {e}")
            raise Exception(f"Error getting message counts: {e}")

    def get_topic_sizes(self, bag_path: str) -> Dict[str, int]:
        """Get total size in bytes for each topic in the bag file using AnyReader"""
        try:
            from rosbags.highlevel import AnyReader
            
            with AnyReader([Path(bag_path)]) as reader:
                # Calculate sizes for each topic
                topic_sizes = {}
                
                for connection in reader.connections:
                    # Sum up message sizes for this topic
                    total_size = 0
                    for (_, _, rawdata) in reader.messages([connection]):
                        total_size += len(rawdata)
                    
                    topic_sizes[connection.topic] = total_size
                
                return topic_sizes
                
        except Exception as e:
            _logger.error(f"Error getting topic sizes: {e}")
            raise Exception(f"Error getting topic sizes: {e}")
    
    def get_topic_stats(self, bag_path: str) -> Dict[str, Dict[str, int]]:
        """Get comprehensive statistics for each topic using AnyReader"""
        try:
            from rosbags.highlevel import AnyReader
            
            with AnyReader([Path(bag_path)]) as reader:
                # Calculate both counts and sizes efficiently
                topic_stats = {}
                
                for connection in reader.connections:
                    count = 0
                    total_size = 0
                    
                    # Process messages for this topic
                    for (_, _, rawdata) in reader.messages([connection]):
                        count += 1
                        total_size += len(rawdata)
                    
                    # Calculate average size
                    avg_size = total_size // count if count > 0 else 0
                    
                    topic_stats[connection.topic] = {
                        'count': count,
                        'size': total_size,
                        'avg_size': avg_size
                    }
                
                return topic_stats
                
        except Exception as e:
            _logger.error(f"Error getting topic stats: {e}")
            raise Exception(f"Error getting topic stats: {e}")
    
    def read_messages(self, bag_path: str, topics: List[str]):
        """
        Read messages from specified topics in the bag file
        
        Args:
            bag_path: Path to bag file
            topics: List of topic names to read from
            
        Yields:
            Tuple of (timestamp, message) where:
            - timestamp: tuple of (seconds, nanoseconds)
            - message: deserialized ROS message
        """
        try:
            from rosbags.highlevel import AnyReader
            
            with AnyReader([Path(bag_path)]) as reader:
                # Pre-filter connections based on selected topics
                selected_connections = [
                    conn for conn in reader.connections 
                    if conn.topic in topics
                ]
                
                if not selected_connections:
                    _logger.warning(f"No matching topics found in {bag_path}")
                    return
                
                # Use AnyReader's high-level message iteration with automatic deserialization
                for (connection, timestamp, rawdata) in reader.messages(connections=selected_connections):
                    try:
                        # Use AnyReader's built-in deserialize method
                        msg = reader.deserialize(rawdata, connection.msgtype)
                        
                        # Convert nanosecond timestamp to (seconds, nanoseconds) format
                        seconds = timestamp // 1_000_000_000
                        nanoseconds = timestamp % 1_000_000_000
                        time_tuple = (int(seconds), int(nanoseconds))
                        
                        yield (time_tuple, msg)
                        
                    except Exception as e:
                        _logger.warning(f"Could not deserialize message for {connection.topic} ({connection.msgtype}): {e}")
                        continue
                
        except Exception as e:
            _logger.error(f"Error reading messages from bag with AnyReader: {e}")
            raise Exception(f"Error reading messages from bag: {e}")


class LegacyBagParser(IBagParser):
    """Legacy rosbag parser implementation as fallback"""
    
    def __init__(self):
        """Initialize legacy parser"""
        _logger.info("Initialized LegacyBagParser for compatibility")
    
    def load_whitelist(self, whitelist_path: str) -> List[str]:
        """Load topics from whitelist file"""
        with open(whitelist_path) as f:
            topics = []
            for line in f.readlines():
                if line.strip() and not line.strip().startswith('#'):
                    topics.append(line.strip())
            return topics
    
    def filter_bag(self, input_bag: str, output_bag: str, topics: List[str], 
                  time_range: Optional[Tuple] = None,
                  progress_callback: Optional[Callable] = None,
                  compression: str = 'none',
                  overwrite: bool = False) -> str:
        """Filter bag using legacy rosbag implementation"""
        try:
            import rosbag
            
            # Check if output file exists
            if os.path.exists(output_bag) and not overwrite:
                raise FileExistsError(f"Output file '{output_bag}' already exists. Use overwrite=True to overwrite.")
            
            # Remove existing file if overwrite is True
            if os.path.exists(output_bag) and overwrite:
                os.remove(output_bag)
            
            start_time = time.time()
            
            # Create output directory if needed
            output_dir = os.path.dirname(output_bag)
            if output_dir:
                os.makedirs(output_dir, exist_ok=True)
            
            # Convert compression format for legacy rosbag
            compression_format = rosbag.Compression.NONE
            if compression == 'bz2':
                compression_format = rosbag.Compression.BZ2
            elif compression == 'lz4':
                compression_format = rosbag.Compression.LZ4
            
            processed_messages = 0
            total_messages = 0
            
            # Count total messages first for progress tracking
            with rosbag.Bag(input_bag, 'r') as input_bag_obj:
                for topic, _, _ in input_bag_obj.read_messages(topics=topics):
                    total_messages += 1
            
            if total_messages == 0:
                _logger.warning(f"No messages found for selected topics in {input_bag}")
                if progress_callback:
                    progress_callback(100)
                return "No messages found for selected topics"
            
            # Filter and write messages
            with rosbag.Bag(input_bag, 'r') as input_bag_obj:
                with rosbag.Bag(output_bag, 'w', compression=compression_format) as output_bag_obj:
                    
                    last_progress = -1
                    
                    for topic, msg, t in input_bag_obj.read_messages(topics=topics):
                        # Check time range if specified
                        if time_range:
                            msg_time = (t.secs, t.nsecs)
                            if (msg_time < time_range[0] or msg_time > time_range[1]):
                                continue
                        
                        output_bag_obj.write(topic, msg, t)
                        processed_messages += 1
                        
                        # Update progress
                        if progress_callback and total_messages > 0:
                            current_progress = int((processed_messages / total_messages) * 100)
                            if current_progress != last_progress:
                                progress_callback(current_progress)
                                last_progress = current_progress
            
            end_time = time.time()
            elapsed = end_time - start_time
            mins, secs = divmod(elapsed, 60)
            
            if progress_callback and last_progress < 100:
                progress_callback(100)
            
            _logger.info(f"Legacy filtered {processed_messages} messages in {elapsed:.2f}s")
            
            return f"Legacy filtering completed in {int(mins)}m {secs:.2f}s"
            
        except Exception as e:
            _logger.error(f"Error filtering bag with legacy parser: {e}")
            raise Exception(f"Error filtering bag: {e}")
    
    def load_bag(self, bag_path: str) -> Tuple[List[str], Dict[str, str], Tuple]:
        """Load bag file using legacy rosbag"""
        try:
            import rosbag
            
            with rosbag.Bag(bag_path, 'r') as bag:
                # Get topics and message types
                info = bag.get_type_and_topic_info()
                topics = list(info.topics.keys())
                connections = {topic: info.topics[topic].msg_type for topic in topics}
                
                # Get time range
                start_time = bag.get_start_time()
                end_time = bag.get_end_time()
                
                # Convert to (seconds, nanoseconds) format
                start = (int(start_time), int((start_time % 1) * 1e9))
                end = (int(end_time), int((end_time % 1) * 1e9))
                
                return topics, connections, (start, end)
                
        except Exception as e:
            _logger.error(f"Error loading bag with legacy parser: {e}")
            raise Exception(f"Error loading bag: {e}")
    
    def inspect_bag(self, bag_path: str) -> str:
        """Inspect bag using legacy rosbag"""
        try:
            import rosbag
            
            topics, connections, (start_time, end_time) = self.load_bag(bag_path)
            topic_stats = self.get_topic_stats(bag_path)
            
            from roseApp.core.util import TimeUtil
            
            def format_size(size_bytes: int) -> str:
                size = float(size_bytes)
                for unit in ['B', 'KB', 'MB', 'GB']:
                    if size < 1024:
                        return f"{size:.1f}{unit}"
                    size /= 1024
                return f"{size:.1f}TB"
            
            result = [f"\nTopics in {bag_path} (Legacy Parser):"]
            result.append("{:<35} {:<35} {:<10} {:<10}".format("Topic", "Message Type", "Count", "Size"))
            result.append("-" * 90)
            
            for topic in topics:
                stats = topic_stats.get(topic, {'count': 0, 'size': 0})
                count = stats['count']
                size = stats['size']
                
                result.append("{:<35} {:<35} {:<10} {:<10}".format(
                    topic[:33], 
                    connections[topic][:33], 
                    count, 
                    format_size(size)
                ))
            
            total_count = sum(stats['count'] for stats in topic_stats.values())
            total_size = sum(stats['size'] for stats in topic_stats.values())
            
            result.append("-" * 90)
            result.append(f"Total: {len(topics)} topics, {total_count} messages, {format_size(total_size)}")
            result.append(f"\nTime range: {TimeUtil.to_datetime(start_time)} - {TimeUtil.to_datetime(end_time)}")
            return "\n".join(result)
            
        except Exception as e:
            _logger.error(f"Error inspecting bag with legacy parser: {e}")
            raise Exception(f"Error inspecting bag: {e}")
    
    def get_message_counts(self, bag_path: str) -> Dict[str, int]:
        """Get message counts using legacy rosbag"""
        try:
            import rosbag
            
            with rosbag.Bag(bag_path, 'r') as bag:
                info = bag.get_type_and_topic_info()
                return {topic: info.topics[topic].message_count for topic in info.topics}
                
        except Exception as e:
            _logger.error(f"Error getting message counts with legacy parser: {e}")
            raise Exception(f"Error getting message counts: {e}")
    
    def get_topic_sizes(self, bag_path: str) -> Dict[str, int]:
        """Get topic sizes using legacy rosbag (estimated)"""
        try:
            import rosbag
            
            topic_sizes = {}
            
            with rosbag.Bag(bag_path, 'r') as bag:
                for topic, msg, _ in bag.read_messages():
                    if topic not in topic_sizes:
                        topic_sizes[topic] = 0
                    # Estimate size (this is approximate for legacy)
                    topic_sizes[topic] += len(str(msg))
            
            return topic_sizes
            
        except Exception as e:
            _logger.error(f"Error getting topic sizes with legacy parser: {e}")
            raise Exception(f"Error getting topic sizes: {e}")
    
    def get_topic_stats(self, bag_path: str) -> Dict[str, Dict[str, int]]:
        """Get topic statistics using legacy rosbag"""
        try:
            message_counts = self.get_message_counts(bag_path)
            topic_sizes = self.get_topic_sizes(bag_path)
            
            stats = {}
            for topic in message_counts:
                count = message_counts[topic]
                size = topic_sizes.get(topic, 0)
                avg_size = size // count if count > 0 else 0
                
                stats[topic] = {
                    'count': count,
                    'size': size,
                    'avg_size': avg_size
                }
            
            return stats
            
        except Exception as e:
            _logger.error(f"Error getting topic stats with legacy parser: {e}")
            raise Exception(f"Error getting topic stats: {e}")
    
    def read_messages(self, bag_path: str, topics: List[str]):
        """Read messages using legacy rosbag"""
        try:
            import rosbag
            
            with rosbag.Bag(bag_path, 'r') as bag:
                for topic, msg, t in bag.read_messages(topics=topics):
                    # Convert timestamp to (seconds, nanoseconds) format
                    time_tuple = (t.secs, t.nsecs)
                    yield (time_tuple, msg)
                    
        except Exception as e:
            _logger.error(f"Error reading messages with legacy parser: {e}")
            raise Exception(f"Error reading messages: {e}")


class ParserHealthChecker:
    """Health checker for parser implementations"""
    
    def __init__(self):
        self._health_cache: Dict[ParserType, ParserHealth] = {}
        self._cache_lock = threading.RLock()
        self._cache_ttl = 300  # 5 minutes
    
    def check_parser_health(self, parser_type: ParserType, force_check: bool = False) -> ParserHealth:
        """Check health of a specific parser type"""
        with self._cache_lock:
            current_time = time.time()
            
            # Return cached result if available and not expired
            if not force_check and parser_type in self._health_cache:
                cached = self._health_cache[parser_type]
                if current_time - cached.last_check < self._cache_ttl:
                    return cached
            
            # Perform health check
            if parser_type == ParserType.ROSBAGS:
                health = self._check_rosbags_health()
            elif parser_type == ParserType.LEGACY:
                health = self._check_legacy_health()
            else:
                health = ParserHealth(
                    parser_type=parser_type,
                    available=False,
                    error_message=f"Unknown parser type: {parser_type}"
                )
            
            health.last_check = current_time
            self._health_cache[parser_type] = health
            
            return health
    
    def _check_rosbags_health(self) -> ParserHealth:
        """Check rosbags parser health"""
        try:
            from rosbags.highlevel import AnyReader
            from rosbags.rosbag1 import Writer as Rosbag1Writer
            
            # Try to get version
            try:
                import rosbags
                version = getattr(rosbags, '__version__', 'unknown')
            except:
                version = 'unknown'
            
            return ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=True,
                version=version,
                performance_score=100.0  # High performance score
            )
            
        except ImportError as e:
            return ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=False,
                error_message=f"rosbags not available: {e}"
            )
        except Exception as e:
            return ParserHealth(
                parser_type=ParserType.ROSBAGS,
                available=False,
                error_message=f"rosbags health check failed: {e}"
            )
    
    def _check_legacy_health(self) -> ParserHealth:
        """Check legacy rosbag parser health"""
        try:
            import rosbag
            
            # Try to get version
            try:
                version = getattr(rosbag, '__version__', 'unknown')
            except:
                version = 'legacy'
            
            return ParserHealth(
                parser_type=ParserType.LEGACY,
                available=True,
                version=version,
                performance_score=30.0  # Lower performance score
            )
            
        except ImportError as e:
            return ParserHealth(
                parser_type=ParserType.LEGACY,
                available=False,
                error_message=f"legacy rosbag not available: {e}"
            )
        except Exception as e:
            return ParserHealth(
                parser_type=ParserType.LEGACY,
                available=False,
                error_message=f"legacy rosbag health check failed: {e}"
            )
    
    def get_best_parser(self) -> ParserType:
        """Get the best available parser based on health and performance"""
        # Check rosbags first (preferred)
        rosbags_health = self.check_parser_health(ParserType.ROSBAGS)
        if rosbags_health.is_healthy():
            return ParserType.ROSBAGS
        
        # Fall back to legacy
        legacy_health = self.check_parser_health(ParserType.LEGACY)
        if legacy_health.is_healthy():
            _logger.warning("Using legacy rosbag parser. Performance may be reduced. Consider installing rosbags for better performance.")
            return ParserType.LEGACY
        
        # No parser available
        raise RuntimeError("No healthy bag parser available. Please install rosbags or legacy rosbag.")
    
    def get_all_health_status(self) -> Dict[ParserType, ParserHealth]:
        """Get health status for all parsers"""
        return {
            ParserType.ROSBAGS: self.check_parser_health(ParserType.ROSBAGS),
            ParserType.LEGACY: self.check_parser_health(ParserType.LEGACY)
        }


# Global health checker
_health_checker = ParserHealthChecker()


def create_parser(parser_type: Optional[ParserType] = None) -> IBagParser:
    """Create parser instance with automatic type selection"""
    if parser_type is None:
        parser_type = _health_checker.get_best_parser()
    
    health = _health_checker.check_parser_health(parser_type)
    if not health.is_healthy():
        raise RuntimeError(f"Parser {parser_type.value} is not healthy: {health.error_message}")
    
    if parser_type == ParserType.ROSBAGS:
        return RosbagsBagParser()
    elif parser_type == ParserType.LEGACY:
        return LegacyBagParser()
    else:
        raise ValueError(f"Unknown parser type: {parser_type}")


def create_best_parser() -> IBagParser:
    """Create the best available parser"""
    best_type = _health_checker.get_best_parser()
    return create_parser(best_type)


def get_parser_health(parser_type: Optional[ParserType] = None) -> ParserHealth:
    """Get health status for a parser type"""
    if parser_type is None:
        parser_type = _health_checker.get_best_parser()
    return _health_checker.check_parser_health(parser_type)


def get_all_parser_health() -> Dict[ParserType, ParserHealth]:
    """Get health status for all parsers"""
    return _health_checker.get_all_health_status()


def check_parser_availability() -> Dict[str, bool]:
    """Check availability of all parsers"""
    health_status = get_all_parser_health()
    return {
        parser_type.value: health.is_healthy()
        for parser_type, health in health_status.items()
    }



"""
Test cases for result_handler module
"""

import pytest
import tempfile
import os
from pathlib import Path
from unittest.mock import Mock, patch

from roseApp.core.result_handler import (
    ResultHandler, OutputFormat, RenderOptions, ExportOptions,
    YAML_AVAILABLE
)


class TestOutputFormat:
    """Test cases for OutputFormat enum"""
    
    def test_output_format_values(self):
        """Test output format enum values"""
        assert OutputFormat.TABLE.value == "table"
        assert OutputFormat.JSON.value == "json"
        assert OutputFormat.YAML.value == "yaml"
        assert OutputFormat.CSV.value == "csv"
        assert OutputFormat.XML.value == "xml"
        assert OutputFormat.HTML.value == "html"
        assert OutputFormat.MARKDOWN.value == "markdown"


class TestRenderOptions:
    """Test cases for RenderOptions class"""
    
    def test_render_options_defaults(self):
        """Test default render options"""
        options = RenderOptions()
        assert options.format == OutputFormat.TABLE
        assert options.verbose is False
        assert options.show_fields is False
        assert options.show_cache_stats is True
        assert options.show_summary is True
        assert options.color is True
        assert options.width is None
        assert options.title is None


class TestExportOptions:
    """Test cases for ExportOptions class"""
    
    def test_export_options_defaults(self):
        """Test default export options"""
        options = ExportOptions()
        assert options.format == OutputFormat.JSON
        assert options.pretty is True
        assert options.include_metadata is True
        assert options.compress is False


class TestResultHandler:
    """Test cases for ResultHandler class"""
    
    def test_handler_initialization(self):
        """Test basic handler initialization"""
        handler = ResultHandler()
        assert handler is not None
    
    def test_prepare_serializable_result(self):
        """Test preparing result for serialization"""
        handler = ResultHandler()
        
        # Test with Path objects
        result = {
            'bag_info': {
                'file_path': Path('/tmp/test.bag'),
                'file_name': 'test.bag'
            },
            'topics': [
                {'name': '/topic1', 'message_count': 100}
            ]
        }
        
        serializable = handler._prepare_serializable_result(result)
        assert serializable['bag_info']['file_path'] == '/tmp/test.bag'
        assert isinstance(serializable['bag_info']['file_path'], str)
    
    def test_format_size(self):
        """Test size formatting"""
        handler = ResultHandler()
        
        assert handler._format_size(500) == "500.0 B"
        assert handler._format_size(1024) == "1.0 KB"
        assert handler._format_size(1024 * 1024) == "1.0 MB"
        assert handler._format_size(1024 * 1024 * 1024) == "1.0 GB"
    
    def test_get_timestamp(self):
        """Test timestamp generation"""
        handler = ResultHandler()
        timestamp = handler._get_timestamp()
        assert isinstance(timestamp, str)
        assert len(timestamp) > 0
    
    def test_render_table(self, sample_analysis_result):
        """Test rendering results as table"""
        handler = ResultHandler()
        
        options = RenderOptions(format=OutputFormat.TABLE)
        result_str = handler.render(sample_analysis_result, options)
        
        # Table rendering outputs to console, returns empty string
        assert result_str == ""
    
    def test_render_list(self, sample_analysis_result):
        """Test rendering results as list"""
        handler = ResultHandler()
        
        options = RenderOptions(format=OutputFormat.LIST)
        result_str = handler.render(sample_analysis_result, options)
        
        # List rendering outputs to console, returns empty string
        assert result_str == ""
    
    def test_render_summary(self, sample_analysis_result):
        """Test rendering results as summary"""
        handler = ResultHandler()
        
        options = RenderOptions(format=OutputFormat.SUMMARY)
        result_str = handler.render(sample_analysis_result, options)
        
        # Summary rendering outputs to console, returns empty string
        assert result_str == ""
    
    def test_render_json(self, sample_analysis_result):
        """Test rendering results as JSON"""
        handler = ResultHandler()
        
        options = RenderOptions(format=OutputFormat.JSON)
        result_str = handler.render(sample_analysis_result, options)
        
        assert isinstance(result_str, str)
        assert '"bag_info"' in result_str
        assert '"topics"' in result_str
    
    def test_render_yaml(self, sample_analysis_result):
        """Test rendering results as YAML"""
        handler = ResultHandler()
        
        if YAML_AVAILABLE:
            options = RenderOptions(format=OutputFormat.YAML)
            result_str = handler.render(sample_analysis_result, options)
            
            assert isinstance(result_str, str)
            assert 'bag_info:' in result_str
        else:
            pytest.skip("YAML library not available")
    
    def test_render_yaml_not_available(self, sample_analysis_result):
        """Test YAML rendering when library not available"""
        with patch('roseApp.core.result_handler.YAML_AVAILABLE', False):
            handler = ResultHandler()
            
            options = RenderOptions(format=OutputFormat.YAML)
            result_str = handler.render(sample_analysis_result, options)
            
            assert result_str == ""
    
    def test_render_markdown(self, sample_analysis_result):
        """Test rendering results as Markdown"""
        handler = ResultHandler()
        
        options = RenderOptions(format=OutputFormat.MARKDOWN)
        result_str = handler.render(sample_analysis_result, options)
        
        assert isinstance(result_str, str)
        assert "# Bag Analysis Report" in result_str
        assert "| Topic | Message Type | Count | Frequency |" in result_str
    
    def test_export_json(self, temp_dir, sample_analysis_result):
        """Test exporting results as JSON"""
        handler = ResultHandler()
        
        output_file = temp_dir / "output.json"
        options = ExportOptions(
            format=OutputFormat.JSON,
            output_file=output_file,
            pretty=True
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is True
        assert output_file.exists()
        
        content = output_file.read_text()
        assert '"bag_info"' in content
        assert '"topics"' in content
    
    def test_export_yaml(self, temp_dir, sample_analysis_result):
        """Test exporting results as YAML"""
        if not YAML_AVAILABLE:
            pytest.skip("YAML library not available")
        
        handler = ResultHandler()
        
        output_file = temp_dir / "output.yaml"
        options = ExportOptions(
            format=OutputFormat.YAML,
            output_file=output_file
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is True
        assert output_file.exists()
        
        content = output_file.read_text()
        assert 'bag_info:' in content
    
    def test_export_csv(self, temp_dir, sample_analysis_result):
        """Test exporting results as CSV"""
        handler = ResultHandler()
        
        output_file = temp_dir / "output.csv"
        options = ExportOptions(
            format=OutputFormat.CSV,
            output_file=output_file
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is True
        assert output_file.exists()
        
        content = output_file.read_text()
        assert 'topic,message_type,message_count,frequency' in content
        assert '/topic1' in content
    
    def test_export_xml(self, temp_dir, sample_analysis_result):
        """Test exporting results as XML"""
        handler = ResultHandler()
        
        output_file = temp_dir / "output.xml"
        options = ExportOptions(
            format=OutputFormat.XML,
            output_file=output_file
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is True
        assert output_file.exists()
        
        content = output_file.read_text()
        assert '<bag_analysis>' in content
        assert '<topics>' in content
    
    def test_export_html(self, temp_dir, sample_analysis_result):
        """Test exporting results as HTML"""
        handler = ResultHandler()
        
        output_file = temp_dir / "output.html"
        options = ExportOptions(
            format=OutputFormat.HTML,
            output_file=output_file
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is True
        assert output_file.exists()
        
        content = output_file.read_text()
        assert '<!DOCTYPE html>' in content
        assert 'ROS Bag Analysis Report' in content
    
    def test_export_markdown(self, temp_dir, sample_analysis_result):
        """Test exporting results as Markdown"""
        handler = ResultHandler()
        
        output_file = temp_dir / "output.md"
        options = ExportOptions(
            format=OutputFormat.MARKDOWN,
            output_file=output_file
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is True
        assert output_file.exists()
        
        content = output_file.read_text()
        assert '# Bag Analysis Report' in content
        assert '| Topic | Message Type | Count | Frequency |' in content
    
    def test_export_unsupported_format(self, sample_analysis_result):
        """Test handling unsupported export format"""
        handler = ResultHandler()
        
        # Create a mock format
        class MockFormat:
            value = "unsupported"
        
        options = ExportOptions(format=MockFormat())
        success = handler.export(sample_analysis_result, options)
        
        assert success is False
    
    def test_export_error_handling(self, temp_dir, sample_analysis_result):
        """Test handling export errors"""
        handler = ResultHandler()
        
        # Use invalid path to trigger error
        output_file = temp_dir / "nonexistent" / "output.json"
        options = ExportOptions(
            format=OutputFormat.JSON,
            output_file=output_file
        )
        
        success = handler.export(sample_analysis_result, options)
        
        assert success is False
    
    def test_convert_format(self, sample_analysis_result):
        """Test format conversion"""
        handler = ResultHandler()
        
        # Test JSON to JSON conversion
        json_str = handler.convert_format(
            sample_analysis_result,
            OutputFormat.JSON,
            OutputFormat.JSON
        )
        
        assert isinstance(json_str, str)
        assert '"bag_info"' in json_str
    
    def test_display_summary(self, sample_analysis_result):
        """Test displaying summary information"""
        handler = ResultHandler()
        
        # This is a private method that outputs to console
        # Just verify it doesn't raise exceptions
        bag_info = sample_analysis_result['bag_info']
        options = RenderOptions()
        
        try:
            handler._display_summary(bag_info, options)
        except Exception as e:
            pytest.fail(f"_display_summary raised exception: {e}")
    
    def test_display_field_analysis(self, sample_analysis_result):
        """Test displaying field analysis"""
        handler = ResultHandler()
        
        field_analysis = sample_analysis_result['field_analysis']
        
        try:
            handler._display_field_analysis(field_analysis)
        except Exception as e:
            pytest.fail(f"_display_field_analysis raised exception: {e}")
    
    def test_display_cache_stats(self, sample_analysis_result):
        """Test displaying cache statistics"""
        handler = ResultHandler()
        
        cache_stats = sample_analysis_result['cache_stats']
        
        try:
            handler._display_cache_stats(cache_stats)
        except Exception as e:
            pytest.fail(f"_display_cache_stats raised exception: {e}")
    
    def test_render_with_custom_console(self):
        """Test handler with custom console"""
        mock_console = Mock()
        handler = ResultHandler(console=mock_console)
        
        assert handler.console == mock_console


# Standard library imports
from pathlib import Path
from typing import Iterable

# Third-party imports
from rich.text import Text
from textual import work
from textual.reactive import reactive
from textual.widgets import DirectoryTree

# Local application imports
from roseApp.tui.components.StatusBar import StatusBar
from roseApp.core.BagManager import BagManager
from roseApp.core.util import get_logger
from roseApp.core.parser import create_parser, ParserType

logger = get_logger("BagExplorer")

class BagExplorer(DirectoryTree):
    """A directory tree widget specialized for selecting ROS bag files"""
    bags = reactive(BagManager(create_parser(ParserType.ROSBAGS)))
    multi_select_mode = reactive(False)
    
    BINDINGS = [
        ("f", "toggle_bags_only", "Filter Bags"),
        ("m", "toggle_multi_select", "Multi Mode"),
    ]
    
    def __init__(self, init_path: str = "."):
        super().__init__(path=init_path)
        self.current_path = Path(init_path)
        self.guide_depth = 2
        self.show_root = True  
        self.show_guides = True
        self.show_only_bags = False
        self.border_title = "File Explorer"
        self.logger = logger.getChild("BagExplorer")
        
        # Use RosbagsBagParser for enhanced performance
        self.bags = BagManager(create_parser(ParserType.ROSBAGS))
        self.logger.info("Using RosbagsBagParser for enhanced performance")

    def on_mount(self) -> None:
        """Initialize when mounted"""
        self.update_border_subtitle()
        self.bags.set_bag_mutate_callback(self.mutate_callback)

    def mutate_callback(self):
        self.mutate_reactive(BagExplorer.bags)
    
    def update_border_subtitle(self):
        """Update subtitle to show multi-select mode status"""
        #mode = "Multi-Select Mode" if self.multi_select_mode else ""
        self.border_subtitle = f" ({self.bags.get_bag_numbers()} selected)" if self.multi_select_mode else ""

    def action_toggle_bags_only(self) -> None:
        self.show_only_bags = not self.show_only_bags
        self.reload()
    
    def action_toggle_multi_select(self) -> None:
        self.multi_select_mode = not self.multi_select_mode
        self.show_only_bags = self.multi_select_mode
        # reset bag manager when switch mode
        self.bags.clear_bags()
        self.reload()   
        self.update_border_subtitle()

    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        """Filter paths based on show_only_bags setting"""
        paths = super().filter_paths(paths)
        paths = [p for p in paths if not p.name.startswith('.')]
        if self.show_only_bags:
            return [p for p in paths if p.is_dir() or p.suffix.lower() == '.bag']
        return paths


    def _handle_directory_selection(self, path: Path, status) -> None:
        """Handle directory selection logic"""
        if path == self.current_path:
            self.path = self.current_path.parent
        else:
            self.path = path
        self.current_path = self.path
        status.update_status(f"Entering directory: {path}")
        

    def _handle_multi_select_bag(self, path: Path, event, status) -> None:
        """Handle bag file selection in multi-select mode"""
        try:
            if self.bags.is_bag_loaded(path):
                self.bags.unload_bag(path)
                event.node.label = Text(path.name)  
                status.update_status(f"Deselected: {path}")
            else:
                self.bags.load_bag(path)
                event.node.label = Text("☑️ ") + Text(path.name)  # Add checkbox symbol
                status.update_status(f"Selected: {path}")   
            self.update_border_subtitle()
        except Exception as e:
            self.logger.error(f"Error loading bag file: {str(e)}", exc_info=True)
            status.update_status(f"Error loading bag file: {str(e)}", "error")
            
    def _handle_single_select_bag(self, path: Path, status) -> None:
        """Handle bag file selection in single-select mode"""
        # for single select mode, clear bags before load current bag
        try:
            self.bags.clear_bags()
            self.bags.load_bag(path)
            status.update_status(f"File: {path} loaded successfully")
        except Exception as e:
            self.logger.error(f"Error loading bag file: {str(e)}", exc_info=True)
            status.update_status(f"Error loading bag file: {str(e)}", "error")

    @work(thread=True)
    async def on_tree_node_selected(self, event: DirectoryTree.NodeSelected) -> None:
        """Handle file selection with support for multi-select mode"""
        path = event.node.data.path
        self.current_node = event.node
        self.show_guides = True
        status = self.app.query_one(StatusBar)

        if path.is_dir():
            self._handle_directory_selection(path, status)
            return

        if not str(path).endswith('.bag'):
            self.bags.clear_bags()
            status.update_status(f"File: {path} is not a bag file", "warning")
            return

        if self.multi_select_mode:
            self._handle_multi_select_bag(path, event, status)
        else:
            self._handle_single_select_bag(path, status)



"""
Test cases for analyzer module
"""

import pytest
import asyncio
from pathlib import Path
from unittest.mock import Mock, patch, AsyncMock

from roseApp.core.analyzer import (
    BagAnalyzer, AnalysisType, BagInfo, MessageTypeInfo, AnalysisResult,
    analyze_bag_async, cleanup_analyzer
)


class TestBagInfo:
    """Test cases for BagInfo class"""
    
    def test_bag_info_creation(self):
        """Test basic BagInfo creation"""
        info = BagInfo(
            path=Path("/tmp/test.bag"),
            size_bytes=1024000,
            topics={"/topic1", "/topic2"},
            message_counts={"/topic1": 100, "/topic2": 50},
            time_range=((123456789, 0), (123456790, 0)),
            connections={"/topic1": "std_msgs/String", "/topic2": "geometry_msgs/Twist"},
            duration_seconds=1.0
        )
        
        assert info.path == Path("/tmp/test.bag")
        assert info.size_bytes == 1024000
        assert len(info.topics) == 2
        assert info.duration_seconds == 1.0


class TestMessageTypeInfo:
    """Test cases for MessageTypeInfo class"""
    
    def test_message_type_info_creation(self):
        """Test basic MessageTypeInfo creation"""
        msg_type = MessageTypeInfo(
            type_name="std_msgs/String",
            fields={"data": {"type": "str"}},
            definition="string data",
            md5sum="992ce8a1687cec8c8bd883ec73ca41d1"
        )
        
        assert msg_type.type_name == "std_msgs/String"
        assert "data" in msg_type.fields
    
    def test_get_field_paths_empty(self):
        """Test getting field paths from empty fields"""
        msg_type = MessageTypeInfo(type_name="std_msgs/Empty")
        paths = msg_type.get_field_paths()
        assert paths == []
    
    def test_get_field_paths_simple(self):
        """Test getting field paths from simple fields"""
        msg_type = MessageTypeInfo(
            type_name="std_msgs/String",
            fields={"data": {"type": "str"}}
        )
        paths = msg_type.get_field_paths()
        assert "data" in paths
    
    def test_get_field_paths_nested(self):
        """Test getting field paths from nested fields"""
        msg_type = MessageTypeInfo(
            type_name="geometry_msgs/Twist",
            fields={
                "linear": {
                    "type": "geometry_msgs/Vector3",
                    "fields": {
                        "x": {"type": "float64"},
                        "y": {"type": "float64"},
                        "z": {"type": "float64"}
                    }
                },
                "angular": {
                    "type": "geometry_msgs/Vector3",
                    "fields": {
                        "x": {"type": "float64"},
                        "y": {"type": "float64"},
                        "z": {"type": "float64"}
                    }
                }
            }
        )
        
        paths = msg_type.get_field_paths()
        assert "linear.x" in paths
        assert "linear.y" in paths
        assert "linear.z" in paths
        assert "angular.x" in paths
        assert "angular.y" in paths
        assert "angular.z" in paths


class TestAnalysisResult:
    """Test cases for AnalysisResult class"""
    
    def test_analysis_result_creation(self):
        """Test basic AnalysisResult creation"""
        bag_info = BagInfo(
            path=Path("/tmp/test.bag"),
            size_bytes=1024000,
            topics={"/topic1"},
            message_counts={"/topic1": 100}
        )
        
        result = AnalysisResult(
            bag_info=bag_info,
            message_types={"std_msgs/String": MessageTypeInfo(type_name="std_msgs/String")},
            analysis_type=AnalysisType.METADATA,
            analysis_time=0.5,
            cached=False
        )
        
        assert result.bag_info == bag_info
        assert result.analysis_type == AnalysisType.METADATA
        assert result.analysis_time == 0.5
        assert not result.cached
    
    def test_get_topic_field_paths(self):
        """Test getting topic field paths"""
        bag_info = BagInfo(
            path=Path("/tmp/test.bag"),
            size_bytes=1024000,
            topics={"/topic1"},
            message_counts={"/topic1": 100},
            connections={"/topic1": "std_msgs/String"}
        )
        
        result = AnalysisResult(
            bag_info=bag_info,
            message_types={"std_msgs/String": MessageTypeInfo(type_name="std_msgs/String")}
        )
        
        paths = result.get_topic_field_paths("/topic1")
        assert paths == []


class TestBagAnalyzer:
    """Test cases for BagAnalyzer class"""
    
    @pytest.fixture
    def mock_parser(self):
        """Create a mock parser"""
        mock = Mock()
        mock.load_bag.return_value = (
            ["/topic1", "/topic2"],
            {"/topic1": "std_msgs/String", "/topic2": "geometry_msgs/Twist"},
            ((123456789, 0), (123456790, 0))
        )
        mock.get_message_counts.return_value = {"/topic1": 100, "/topic2": 50}
        return mock
    
    def test_analyzer_initialization(self):
        """Test basic analyzer initialization"""
        analyzer = BagAnalyzer(max_workers=2)
        assert analyzer.max_workers == 2
    
    @pytest.mark.asyncio
    async def test_analyze_bag_async_metadata(self, temp_dir, mock_parser):
        """Test async bag analysis with metadata type"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('roseApp.core.parser.create_best_parser', return_value=mock_parser):
            with patch.object(bag_path, 'stat') as mock_stat:
                mock_stat.return_value.st_size = 1024000
                mock_stat.return_value.st_mtime = 1234567890
                
                analyzer = BagAnalyzer()
                result = await analyzer.analyze_bag_async(
                    bag_path=bag_path,
                    analysis_type=AnalysisType.METADATA
                )
                
                assert isinstance(result, AnalysisResult)
                assert result.analysis_type == AnalysisType.METADATA
                assert result.bag_info.path == bag_path
                assert len(result.bag_info.topics) == 2
                assert result.bag_info.message_counts["/topic1"] == 100
    
    @pytest.mark.asyncio
    async def test_analyze_bag_async_full_analysis(self, temp_dir, mock_parser):
        """Test async bag analysis with full analysis type"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('roseApp.core.parser.create_best_parser', return_value=mock_parser):
            with patch.object(bag_path, 'stat') as mock_stat:
                mock_stat.return_value.st_size = 1024000
                mock_stat.return_value.st_mtime = 1234567890
                
                # Mock the message reading
                def mock_read_messages(bag_path, topics):
                    # Mock message objects
                    class MockMessage:
                        def __init__(self):
                            self.data = "test"
                    
                    yield ((123456789, 0), MockMessage())
                
                mock_parser.read_messages = mock_read_messages
                
                analyzer = BagAnalyzer()
                result = await analyzer.analyze_bag_async(
                    bag_path=bag_path,
                    analysis_type=AnalysisType.FULL_ANALYSIS
                )
                
                assert isinstance(result, AnalysisResult)
                assert result.analysis_type == AnalysisType.FULL_ANALYSIS
                assert len(result.message_types) > 0
    
    @pytest.mark.asyncio
    async def test_analyze_bag_async_with_progress(self, temp_dir, mock_parser):
        """Test async bag analysis with progress callback"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        progress_values = []
        
        def progress_callback(progress):
            progress_values.append(progress)
        
        with patch('roseApp.core.parser.create_best_parser', return_value=mock_parser):
            with patch.object(bag_path, 'stat') as mock_stat:
                mock_stat.return_value.st_size = 1024000
                mock_stat.return_value.st_mtime = 1234567890
                
                analyzer = BagAnalyzer()
                result = await analyzer.analyze_bag_async(
                    bag_path=bag_path,
                    analysis_type=AnalysisType.METADATA,
                    progress_callback=progress_callback
                )
                
                assert isinstance(result, AnalysisResult)
                assert len(progress_values) > 0
                assert 100.0 in progress_values
    
    @pytest.mark.asyncio
    async def test_analyze_bag_async_with_cache(self, temp_dir, mock_parser):
        """Test async bag analysis with caching"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('roseApp.core.parser.create_best_parser', return_value=mock_parser):
            with patch.object(bag_path, 'stat') as mock_stat:
                mock_stat.return_value.st_size = 1024000
                mock_stat.return_value.st_mtime = 1234567890
                
                # Mock cache to return cached result
                mock_cache = Mock()
                mock_cache.get.return_value = AnalysisResult(
                    bag_info=BagInfo(
                        path=bag_path,
                        size_bytes=1024000,
                        topics={"/topic1"},
                        message_counts={"/topic1": 100}
                    ),
                    cached=True
                )
                
                analyzer = BagAnalyzer()
                analyzer.cache = mock_cache
                
                result = await analyzer.analyze_bag_async(
                    bag_path=bag_path,
                    analysis_type=AnalysisType.METADATA
                )
                
                assert isinstance(result, AnalysisResult)
                assert result.cached
    
    @pytest.mark.asyncio
    async def test_analyze_bag_async_error_handling(self, temp_dir):
        """Test error handling in async bag analysis"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('roseApp.core.parser.create_best_parser', side_effect=Exception("Parser error")):
            analyzer = BagAnalyzer()
            result = await analyzer.analyze_bag_async(
                bag_path=bag_path,
                analysis_type=AnalysisType.METADATA
            )
            
            assert isinstance(result, AnalysisResult)
            assert len(result.errors) > 0
            assert "Parser error" in result.errors[0]
    
    def test_extract_message_fields(self):
        """Test message field extraction"""
        analyzer = BagAnalyzer()
        
        # Mock message with __slots__
        class MockMessage:
            __slots__ = ['data', 'header']
            
            def __init__(self):
                self.data = "test"
                self.header = Mock()
                self.header.stamp = Mock()
                self.header.stamp.secs = 123456789
                self.header.stamp.nsecs = 0
        
        message = MockMessage()
        fields = analyzer._extract_message_fields(message)
        
        assert "data" in fields
        assert "header" in fields
        assert fields["data"]["type"] == "str"
    
    def test_analyze_field_value(self):
        """Test field value analysis"""
        analyzer = BagAnalyzer()
        
        # Test primitive types
        result = analyzer._analyze_field_value(42)
        assert result["type"] == "int"
        assert result["value_sample"] == "42"
        
        # Test list
        result = analyzer._analyze_field_value([1, 2, 3])
        assert result["type"] == "list"
        assert result["array"] is True
        assert result["length"] == 3
    
    def test_merge_fields(self):
        """Test field merging"""
        analyzer = BagAnalyzer()
        
        existing = {
            "field1": {"type": "int"},
            "field2": {"type": "str"}
        }
        
        new = {
            "field2": {"type": "str", "value": "test"},
            "field3": {"type": "float"}
        }
        
        merged = analyzer._merge_fields(existing, new)
        assert len(merged) == 3
        assert "field1" in merged
        assert "field2" in merged
        assert "field3" in merged
    
    def test_cleanup(self):
        """Test analyzer cleanup"""
        analyzer = BagAnalyzer()
        analyzer.cleanup()
        # Should not raise any exceptions


class TestGlobalFunctions:
    """Test cases for global functions"""
    
    @pytest.mark.asyncio
    async def test_analyze_bag_async_global(self, temp_dir, mock_parser):
        """Test global analyze_bag_async function"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('roseApp.core.parser.create_best_parser', return_value=mock_parser):
            with patch.object(bag_path, 'stat') as mock_stat:
                mock_stat.return_value.st_size = 1024000
                mock_stat.return_value.st_mtime = 1234567890
                
                result = await analyze_bag_async(
                    bag_path=bag_path,
                    analysis_type=AnalysisType.METADATA
                )
                
                assert isinstance(result, AnalysisResult)
    
    def test_cleanup_analyzer(self):
        """Test global cleanup function"""
        cleanup_analyzer()
        # Should not raise any exceptions


"""
Async bag analyzer with intelligent caching and message type analysis
"""

import asyncio
import time
from pathlib import Path
from typing import Optional, Callable, Dict, Set, List, Any
from dataclasses import dataclass, field
from enum import Enum
from concurrent.futures import ThreadPoolExecutor

from .parser import create_best_parser
from .cache import get_cache
from .util import get_logger


class AnalysisType(Enum):
    """Types of bag analysis"""
    METADATA = "metadata"
    FULL_ANALYSIS = "full_analysis"


@dataclass
class BagInfo:
    """Bag file information"""
    path: Path
    size_bytes: int
    topics: Set[str]
    message_counts: Dict[str, int]
    time_range: Optional[tuple] = None
    connections: Dict[str, str] = field(default_factory=dict)
    duration_seconds: float = 0.0


@dataclass
class MessageTypeInfo:
    """Message type information"""
    type_name: str
    fields: Dict[str, Any] = field(default_factory=dict)
    definition: str = ""
    md5sum: str = ""

    def get_field_paths(self) -> List[str]:
        """Get all field paths for this message type"""
        paths = []
        
        def extract_paths(fields_dict, prefix=""):
            for field_name, field_info in fields_dict.items():
                current_path = f"{prefix}.{field_name}" if prefix else field_name
                paths.append(current_path)
                
                if isinstance(field_info, dict) and 'fields' in field_info:
                    extract_paths(field_info['fields'], current_path)
        
        extract_paths(self.fields)
        return paths


@dataclass
class AnalysisResult:
    """Complete analysis result"""
    bag_info: BagInfo
    message_types: Dict[str, MessageTypeInfo] = field(default_factory=dict)
    analysis_type: AnalysisType = AnalysisType.METADATA
    analysis_time: float = 0.0
    cached: bool = False
    errors: List[str] = field(default_factory=list)

    def get_topic_field_paths(self, topic: str) -> List[str]:
        """Get field paths for a specific topic"""
        msg_type = self.bag_info.connections.get(topic)
        if msg_type and msg_type in self.message_types:
            return self.message_types[msg_type].get_field_paths()
        return []


class BagAnalyzer:
    """Async bag analyzer with intelligent caching"""
    
    def __init__(self, max_workers: int = 4):
        self.max_workers = max_workers
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        self.cache = get_cache()
        self.logger = get_logger()
        
    async def analyze_bag_async(
        self,
        bag_path: Path,
        analysis_type: AnalysisType = AnalysisType.METADATA,
        progress_callback: Optional[Callable[[float], None]] = None
    ) -> AnalysisResult:
        """Analyze bag file asynchronously"""
        start_time = time.time()
        
        # Generate cache key
        cache_key = f"analysis_{bag_path}_{analysis_type.value}_{bag_path.stat().st_mtime}"
        
        # Check cache first
        cached_result = self.cache.get(cache_key)
        if cached_result:
            self.logger.info(f"Using cached analysis for {bag_path}")
            cached_result.cached = True
            return cached_result
        
        try:
            if progress_callback:
                progress_callback(10.0)
            
            # Create parser
            parser = create_best_parser()
            
            if progress_callback:
                progress_callback(20.0)
            
            # Load basic bag info
            bag_info = await self._load_bag_info_async(parser, bag_path)
            
            if progress_callback:
                progress_callback(60.0)
            
            # Create result
            result = AnalysisResult(
                bag_info=bag_info,
                analysis_type=analysis_type,
                analysis_time=time.time() - start_time,
                cached=False
            )
            
            # Perform message type analysis if requested
            if analysis_type == AnalysisType.FULL_ANALYSIS:
                result.message_types = await self._analyze_message_types_async(
                    bag_info.connections,
                    bag_path
                )
            
            if progress_callback:
                progress_callback(100.0)
            
            # Cache the result
            self.cache.put(cache_key, result, ttl=3600)  # Cache for 1 hour
            
            self.logger.info(f"Analysis completed in {result.analysis_time:.2f}s")
            return result
            
        except Exception as e:
            self.logger.error(f"Analysis failed: {e}")
            # Return minimal result on error
            return AnalysisResult(
                bag_info=BagInfo(
                    path=bag_path,
                    size_bytes=bag_path.stat().st_size if bag_path.exists() else 0,
                    topics=set(),
                    message_counts={}
                ),
                analysis_type=analysis_type,
                analysis_time=time.time() - start_time,
                cached=False,
                errors=[str(e)]
            )
    
    async def _load_bag_info_async(self, parser, bag_path: Path) -> BagInfo:
        """Load basic bag information asynchronously"""
        loop = asyncio.get_event_loop()
        
        # Run parser operations in thread pool
        topics, connections, time_range = await loop.run_in_executor(
            self.executor,
            parser.load_bag,
            str(bag_path)
        )
        
        message_counts = await loop.run_in_executor(
            self.executor,
            parser.get_message_counts,
            str(bag_path)
        )
        
        # Calculate duration
        duration = 0.0
        if time_range and len(time_range) >= 2:
            # time_range contains tuples of (seconds, nanoseconds)
            start_time = time_range[0]
            end_time = time_range[1]
            
            if isinstance(start_time, tuple) and isinstance(end_time, tuple):
                # Convert to seconds with nanosecond precision
                start_seconds = start_time[0] + start_time[1] / 1e9
                end_seconds = end_time[0] + end_time[1] / 1e9
                duration = end_seconds - start_seconds
            else:
                # Fallback for simple numeric timestamps
                try:
                    duration = float(end_time) - float(start_time)
                except (TypeError, ValueError):
                    duration = 0.0
        
        return BagInfo(
            path=bag_path,
            size_bytes=bag_path.stat().st_size,
            topics=set(topics),
            message_counts=message_counts,
            time_range=time_range,
            connections=connections,
            duration_seconds=duration
        )
    
    async def _analyze_message_types_async(
        self,
        connections: Dict[str, str],
        bag_path: Path
    ) -> Dict[str, MessageTypeInfo]:
        """Analyze message types by sampling messages"""
        message_types = {}
        
        # Get unique message types and their topics
        type_to_topics = {}
        for topic, msg_type in connections.items():
            if msg_type not in type_to_topics:
                type_to_topics[msg_type] = []
            type_to_topics[msg_type].append(topic)
        
        loop = asyncio.get_event_loop()
        parser = create_best_parser()
        
        for msg_type, topics in type_to_topics.items():
            try:
                # Sample messages from the first topic of this type
                sample_topic = topics[0]
                
                # Read a few sample messages to analyze structure
                def _sample_messages():
                    try:
                        fields = {}
                        sample_count = 0
                        max_samples = 3  # Limit samples for performance
                        
                        message_generator = parser.read_messages(str(bag_path), [sample_topic])
                        if message_generator is not None:
                            for timestamp, message in message_generator:
                                 if sample_count >= max_samples:
                                     break
                                 
                                 # Analyze message structure
                                 message_fields = self._extract_message_fields(message)
                                 fields = self._merge_fields(fields, message_fields)
                                 sample_count += 1
                        
                        return fields
                    except Exception as e:
                        self.logger.debug(f"Could not sample messages for {msg_type}: {e}")
                        return {}
                
                # Run in executor to avoid blocking
                fields = await loop.run_in_executor(self.executor, _sample_messages)
                
                message_types[msg_type] = MessageTypeInfo(
                    type_name=msg_type,
                    fields=fields,
                    definition="",  # Could be populated from message definition
                    md5sum=""  # Could be calculated from definition
                )
                
            except Exception as e:
                self.logger.warning(f"Failed to analyze message type {msg_type}: {e}")
                # Create empty message type info as fallback
                message_types[msg_type] = MessageTypeInfo(
                    type_name=msg_type,
                    fields={},
                    definition="",
                    md5sum=""
                )
        
        return message_types
    
    def _extract_message_fields(self, message) -> Dict[str, Any]:
        """Extract field structure from a ROS message"""
        fields = {}
        
        if hasattr(message, '__slots__'):
            # ROS message with __slots__
            for field_name in message.__slots__:
                if hasattr(message, field_name):
                    field_value = getattr(message, field_name)
                    fields[field_name] = self._analyze_field_value(field_value)
        elif hasattr(message, '__dict__'):
            # Regular object with __dict__
            for field_name, field_value in message.__dict__.items():
                if not field_name.startswith('_'):
                    fields[field_name] = self._analyze_field_value(field_value)
        else:
            # Try to get common ROS message fields
            common_fields = ['header', 'data', 'pose', 'twist', 'position', 'orientation']
            for field_name in common_fields:
                if hasattr(message, field_name):
                    field_value = getattr(message, field_name)
                    fields[field_name] = self._analyze_field_value(field_value)
        
        return fields
    
    def _analyze_field_value(self, value) -> Dict[str, Any]:
        """Analyze the type and structure of a field value"""
        field_info = {
            'type': type(value).__name__,
            'value_sample': None
        }
        
        # Handle different value types
        if hasattr(value, '__slots__') or hasattr(value, '__dict__'):
            # Nested message
            field_info['fields'] = self._extract_message_fields(value)
        elif isinstance(value, (list, tuple)) and len(value) > 0:
            # Array/list
            field_info['array'] = True
            field_info['length'] = len(value)
            field_info['element_type'] = type(value[0]).__name__
            # Analyze first element if it's a complex type
            if hasattr(value[0], '__slots__') or hasattr(value[0], '__dict__'):
                field_info['element_fields'] = self._extract_message_fields(value[0])
        elif isinstance(value, (int, float, str, bool)):
            # Primitive type
            field_info['value_sample'] = str(value)[:50]  # Limit sample length
        
        return field_info
    
    def _merge_fields(self, existing_fields: Dict[str, Any], new_fields: Dict[str, Any]) -> Dict[str, Any]:
        """Merge field structures from multiple message samples"""
        merged = existing_fields.copy()
        
        for field_name, field_info in new_fields.items():
            if field_name not in merged:
                merged[field_name] = field_info
            else:
                # Merge nested fields if present
                if 'fields' in field_info and 'fields' in merged[field_name]:
                    merged[field_name]['fields'] = self._merge_fields(
                        merged[field_name]['fields'], 
                        field_info['fields']
                    )
        
        return merged
    
    def cleanup(self):
        """Clean up resources"""
        self.executor.shutdown(wait=True)


# Global analyzer instance
_analyzer = None


async def analyze_bag_async(
    bag_path: Path,
    analysis_type: AnalysisType = AnalysisType.METADATA,
    progress_callback: Optional[Callable[[float], None]] = None
) -> AnalysisResult:
    """Analyze bag file asynchronously - main public interface"""
    global _analyzer
    
    if _analyzer is None:
        _analyzer = BagAnalyzer()
    
    return await _analyzer.analyze_bag_async(
        bag_path=bag_path,
        analysis_type=analysis_type,
        progress_callback=progress_callback
    )


def cleanup_analyzer():
    """Clean up global analyzer"""
    global _analyzer
    if _analyzer:
        _analyzer.cleanup()
        _analyzer = None 


# Standard library imports
from collections import defaultdict
from pathlib import Path
from typing import Iterable, Optional

# Third-party imports
from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Container
from textual.fuzzy import FuzzySearch
from textual.widgets import Input, Tree

# Local application imports
from roseApp.tui.components.BagExplorer import BagExplorer
from roseApp.core.BagManager import BagManager
from roseApp.core.util import get_logger

logger = get_logger("TopicPanel")



class TopicTree(Tree):
    """A tree widget for displaying ROS bag topics with multi-selection capability"""
    
    def __init__(self):
        super().__init__("Topics")
        self.border_subtitle = "Selected: 0"
        self.fuzzy_searcher = FuzzySearch(case_sensitive=False)
        self.multi_select_mode = False
        self.show_root = False
        self._search_input = None


    def on_mount(self) -> None:
        """Initialize when mounted"""
        super().on_mount()
        self.root.expand()
        self._search_input = self.parent.query_one(TopicSearchInput)
        self.watch(self.app.query_one(BagExplorer), "bags", self.handle_bags_change)
        self.watch(self.app.query_one(BagExplorer), "multi_select_mode", 
                                self.handle_multi_select_mode_change)

    @property
    def bags(self) -> BagManager:
        return self.app.query_one(BagExplorer).bags
    

    def handle_bags_change(self, bags: BagManager) -> None:
        """Handle changes in BagManager and update topics accordingly"""        
        self.render_topics()

    def handle_multi_select_mode_change(self, multi_select_mode: bool) -> None:
        """Handle multi select mode change"""
        self.multi_select_mode = multi_select_mode

    def get_node_label(self, topic: str, selected: bool = False) -> Text:
        """
        Get the label for a topic node based on mode and selection state.
        
        Args:
            topic: The topic name
            selected: Whether the topic is selected
        """
        if self.multi_select_mode:
            # Get count from bag manager
            count = self.bags.get_topic_summary().get(topic, 0)
            label = f"{topic} [{count}]"
        else:
            label = topic
            
        if selected:
            return Text("√ ") + Text(label)
        return Text(label)

    def filter_topics(self,topics,search_text: str) -> None:
        """Filter topics based on search text using fuzzy search.
        Do not change data in bag manager, only for display
        """
        self.root.remove_children()
        
        if not search_text:
            filtered_topics = topics
        else:
            scored_topics = [
                (topic, self.fuzzy_searcher.match(search_text, topic)[0])
                for topic in topics
            ]
            filtered_topics = [
                topic for topic, score in sorted(
                    scored_topics,
                    key=lambda x: x[1],
                    reverse=True
                ) if score > 0
            ]
        
        return filtered_topics
        

    def render_topics(self) -> None:
        """Set topics based on BagManager's state"""


        topics = list(self.bags.get_topic_summary().keys())
        search_text = self._search_input.value
        
        filtered_topics = self.filter_topics(topics,search_text)
        
        
        
        self.root.remove_children()
        for topic in filtered_topics:
            is_selected = topic in self.bags.get_selected_topics()
            self.root.add(
                self.get_node_label(topic,is_selected),
                data={"topic": topic, "selected": is_selected},
                allow_expand=False
            )
        
        self.update_border_subtitle()
        self.update_border_title()

    def update_border_subtitle(self):
        """Update subtitle with selected topics count"""
        self.app.query_one(TopicTreePanel).border_subtitle = f"Topic selected: {len(self.bags.get_selected_topics())}"
    
    def update_border_title(self):
        if self.multi_select_mode:
            self.app.query_one(TopicTreePanel).border_title = "Topics(Show Common Topics)"
        else:
            self.app.query_one(TopicTreePanel).border_title = "Topics"

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        """Handle topic selection toggle"""
        if event.node.allow_expand:
            return
            
        data = event.node.data
        if data:
            data["selected"] = not data["selected"]
            topic = data["topic"]
            
            if data["selected"]:
                self.bags.select_topic(topic)
            else:
                self.bags.deselect_topic(topic)
            
            print(self.bags.get_selected_topics())
            event.node.label = self.get_node_label(topic, data["selected"])
            self.update_border_subtitle()



    def toggle_select_all(self) -> 'tuple[bool, int]':
        """
        Toggle selection state of all topics.
        
        Returns:
            tuple[bool, int]: (is_all_deselected, count_of_selected)
        """
        if not self.root.children:
            return False, 0
            
        # Check if all topics are already selected
        all_selected = all(node.data["selected"] for node in self.root.children)
        
        # Toggle selection state
        for node in self.root.children:
            node.data["selected"] = not all_selected
            topic = node.data.get("topic")
            if node.data["selected"]:
                self.bags.select_topic(topic)
                node.label = self.get_node_label(topic, True)
            else:
                self.bags.deselect_topic(topic)
                node.label = self.get_node_label(topic, False)
        
        print(self.bags)
        self.update_border_subtitle()
        return all_selected, len(self.bags.get_selected_topics())

class TopicSearchInput(Input):
    """Input widget for searching topics"""
    
    def __init__(self):
        super().__init__(placeholder="Search topics...", id="topic-search")
        self._topic_tree = None

    def on_mount(self) -> None:
        """Get reference to topic tree when mounted"""
        self._topic_tree = self.parent.query_one(TopicTree)

    def on_input_changed(self, event: Input.Changed) -> None:
        """Handle search input changes"""
        self._topic_tree.render_topics()
            

class TopicTreePanel(Container):
    """A wrapper component that contains a search input and a topic tree"""
    
    def __init__(self):
        super().__init__()

    def compose(self) -> ComposeResult:
        """Create child widgets"""
        yield TopicSearchInput()
        yield TopicTree()

    def on_mount(self) -> None:
        """Initialize when mounted"""
        self.border_title = "Topics"

    def get_topic_tree(self) -> TopicTree:
        """Get the topic tree"""
        return self.query_one(TopicTree)
    




"""
Test cases for BagManager module
"""

import pytest
from pathlib import Path
from unittest.mock import Mock, patch

from roseApp.core.BagManager import (
    BagManager, BagStatus, Bag, BagInfo, FilterConfig, CompressionType
)


class TestBagInfo:
    """Test cases for BagInfo class"""
    
    def test_bag_info_creation(self):
        """Test basic BagInfo creation"""
        info = BagInfo(
            time_range=((123456789, 0), (123456790, 0)),
            init_time_range=((123456789, 0), (123456790, 0)),
            size=1024000,
            topics={"/topic1", "/topic2"},
            size_after_filter=1024000
        )
        
        assert info.size == 1024000
        assert len(info.topics) == 2
        assert info.size_str == "976.56KB"
    
    def test_time_range_str(self):
        """Test time range string formatting"""
        info = BagInfo(
            time_range=((123456789, 0), (123456790, 0)),
            init_time_range=((123456789, 0), (123456790, 0)),
            size=1000,
            topics={"/topic1"},
            size_after_filter=1000
        )
        
        start_str, end_str = info.time_range_str
        assert "09/02/73 03:46:29" in start_str
        assert "09/02/73 03:46:30" in end_str
    
    def test_size_formatting(self):
        """Test size formatting in different units"""
        info = BagInfo(
            time_range=((0, 0), (1, 0)),
            init_time_range=((0, 0), (1, 0)),
            size=500,
            topics=set(),
            size_after_filter=500
        )
        assert info.size_str == "500.00B"
        
        info.size = 1024 * 1024
        assert info.size_str == "1.00MB"


class TestBag:
    """Test cases for Bag class"""
    
    def test_bag_creation(self):
        """Test basic Bag creation"""
        path = Path("/tmp/test.bag")
        info = BagInfo(
            time_range=((0, 0), (1, 0)),
            init_time_range=((0, 0), (1, 0)),
            size=1000,
            topics={"/topic1"},
            size_after_filter=1000
        )
        
        bag = Bag(path, info)
        assert bag.path == path
        assert bag.status == BagStatus.IDLE
        assert len(bag.selected_topics) == 0
    
    def test_bag_topic_selection(self):
        """Test topic selection functionality"""
        path = Path("/tmp/test.bag")
        info = BagInfo(
            time_range=((0, 0), (1, 0)),
            init_time_range=((0, 0), (1, 0)),
            size=1000,
            topics={"/topic1", "/topic2"},
            size_after_filter=1000
        )
        
        bag = Bag(path, info)
        bag.set_selected_topics({"/topic1"})
        assert len(bag.selected_topics) == 1
        assert "/topic1" in bag.selected_topics
    
    def test_filter_config_generation(self):
        """Test filter config generation"""
        path = Path("/tmp/test.bag")
        info = BagInfo(
            time_range=((0, 0), (1, 0)),
            init_time_range=((0, 0), (1, 0)),
            size=1000,
            topics={"/topic1", "/topic2"},
            size_after_filter=1000
        )
        
        bag = Bag(path, info)
        bag.set_selected_topics({"/topic1"})
        
        config = bag.get_filter_config()
        assert config.compression == "none"
        assert len(config.topic_list) == 1
        assert "/topic1" in config.topic_list
    
    def test_bag_status_management(self):
        """Test bag status management"""
        path = Path("/tmp/test.bag")
        info = BagInfo(
            time_range=((0, 0), (1, 0)),
            init_time_range=((0, 0), (1, 0)),
            size=1000,
            topics=set(),
            size_after_filter=1000
        )
        
        bag = Bag(path, info)
        assert bag.status == BagStatus.IDLE
        
        bag.set_status(BagStatus.SUCCESS)
        assert bag.status == BagStatus.SUCCESS
        
        bag.set_status(BagStatus.ERROR)
        assert bag.status == BagStatus.ERROR


class TestBagManager:
    """Test cases for BagManager class"""
    
    def test_bag_manager_initialization(self):
        """Test basic BagManager initialization"""
        manager = BagManager()
        assert len(manager.bags) == 0
        assert len(manager.selected_topics) == 0
        assert manager.compression == CompressionType.NONE.value
    
    def test_bag_manager_with_parser(self, mock_parser):
        """Test BagManager initialization with custom parser"""
        manager = BagManager(parser=mock_parser)
        assert manager._parser == mock_parser
    
    def test_load_bag(self, temp_dir, mock_parser):
        """Test loading a bag file"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag_path)
            
            assert len(manager.bags) == 1
            assert bag_path in manager.bags
            assert len(manager.selected_topics) == 0
    
    def test_load_duplicate_bag(self, temp_dir, mock_parser):
        """Test loading duplicate bag file raises error"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag_path)
            
            with pytest.raises(ValueError, match="already exists"):
                manager.load_bag(bag_path)
    
    def test_unload_bag(self, temp_dir, mock_parser):
        """Test unloading a bag file"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag_path)
            assert len(manager.bags) == 1
            
            manager.unload_bag(bag_path)
            assert len(manager.bags) == 0
    
    def test_unload_nonexistent_bag(self, temp_dir):
        """Test unloading nonexistent bag raises error"""
        bag_path = temp_dir / "nonexistent.bag"
        manager = BagManager()
        
        with pytest.raises(KeyError, match="not found"):
            manager.unload_bag(bag_path)
    
    def test_clear_bags(self, temp_dir, mock_parser):
        """Test clearing all bags"""
        bag1_path = temp_dir / "test1.bag"
        bag2_path = temp_dir / "test2.bag"
        bag1_path.touch()
        bag2_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag1_path)
            manager.load_bag(bag2_path)
            assert len(manager.bags) == 2
            
            manager.clear_bags()
            assert len(manager.bags) == 0
    
    def test_topic_selection(self, temp_dir, mock_parser):
        """Test topic selection functionality"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag_path)
            
            # Test topic selection
            manager.select_topic("/topic1")
            assert "/topic1" in manager.selected_topics
            
            # Test topic deselection
            manager.deselect_topic("/topic1")
            assert "/topic1" not in manager.selected_topics
            
            # Test clear all topics
            manager.select_topic("/topic1")
            manager.select_topic("/topic2")
            assert len(manager.selected_topics) == 2
            
            manager.clear_selected_topics()
            assert len(manager.selected_topics) == 0
    
    def test_get_common_topics(self, temp_dir, mock_parser):
        """Test getting common topics across bags"""
        bag1_path = temp_dir / "test1.bag"
        bag2_path = temp_dir / "test2.bag"
        bag1_path.touch()
        bag2_path.touch()
        
        def mock_load_bag(path):
            if "test1" in str(path):
                return ["/topic1", "/topic2"], {}, ((0, 0), (1, 0))
            else:
                return ["/topic2", "/topic3"], {}, ((0, 0), (1, 0))
        
        mock_parser.load_bag.side_effect = mock_load_bag
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag1_path)
            manager.load_bag(bag2_path)
            
            common_topics = manager.get_common_topics()
            assert "/topic2" in common_topics
            assert len(common_topics) == 1
    
    def test_get_topic_summary(self, temp_dir, mock_parser):
        """Test getting topic summary"""
        bag1_path = temp_dir / "test1.bag"
        bag2_path = temp_dir / "test2.bag"
        bag1_path.touch()
        bag2_path.touch()
        
        def mock_load_bag(path):
            if "test1" in str(path):
                return ["/topic1", "/topic2"], {}, ((0, 0), (1, 0))
            else:
                return ["/topic2", "/topic3"], {}, ((0, 0), (1, 0))
        
        mock_parser.load_bag.side_effect = mock_load_bag
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(bag1_path)
            manager.load_bag(bag2_path)
            
            summary = manager.get_topic_summary()
            assert summary["/topic1"] == 1
            assert summary["/topic2"] == 2
            assert summary["/topic3"] == 1
    
    def test_compression_type_management(self):
        """Test compression type management"""
        manager = BagManager()
        
        # Test valid compression types
        manager.set_compression_type("bz2")
        assert manager.get_compression_type() == "bz2"
        
        manager.set_compression_type("lz4")
        assert manager.get_compression_type() == "lz4"
        
        manager.set_compression_type("none")
        assert manager.get_compression_type() == "none"
        
        # Test invalid compression type
        with pytest.raises(ValueError):
            manager.set_compression_type("invalid")
    
    def test_callbacks(self, temp_dir, mock_parser):
        """Test callback functionality"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        callback_called = False
        
        def bag_mutate_callback():
            nonlocal callback_called
            callback_called = True
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.set_bag_mutate_callback(bag_mutate_callback)
            
            manager.load_bag(bag_path)
            assert callback_called
    
    def test_get_single_bag(self, temp_dir, mock_parser):
        """Test getting single bag"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            assert manager.get_single_bag() is None
            
            manager.load_bag(bag_path)
            single_bag = manager.get_single_bag()
            assert single_bag is not None
            assert single_bag.path == bag_path
    
    def test_is_bag_loaded(self, temp_dir, mock_parser):
        """Test checking if bag is loaded"""
        bag_path = temp_dir / "test.bag"
        bag_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            assert not manager.is_bag_loaded(bag_path)
            
            manager.load_bag(bag_path)
            assert manager.is_bag_loaded(bag_path)
    
    def test_filter_bag(self, temp_dir, mock_parser):
        """Test filtering a bag"""
        input_path = temp_dir / "input.bag"
        output_path = temp_dir / "output.bag"
        input_path.touch()
        output_path.touch()
        
        with patch('pathlib.Path.stat') as mock_stat:
            mock_stat.return_value.st_size = 1024000
            
            manager = BagManager(parser=mock_parser)
            manager.load_bag(input_path)
            
            config = FilterConfig(
                time_range=((0, 0), (1, 0)),
                topic_list=["/topic1"],
                compression="bz2"
            )
            
            manager.filter_bag(input_path, config, output_path)
            
            assert manager.bags[input_path].status == BagStatus.SUCCESS
    
    def test_parser_type_detection(self, mock_parser):
        """Test parser type detection"""
        with patch.object(mock_parser, '__class__') as mock_class:
            mock_class.__name__ = 'RosbagsBagParser'
            manager = BagManager(parser=mock_parser)
            assert manager.get_parser_type() == 'rosbags'